/***************************************************************************
	Read COPYRIGHT file in this directory
****************************************************************************/


#include "gamma.h"
#include "gamma_leading.h"
static char *gamma_authors = "ciaccio@disi.unige.it";


/************************************************************************
 *  GAMMA Data structures						*
 ************************************************************************/

/*
 * No. and array of NICs probed by the GAMMA driver.
 * Initialized by the driver probe function.
 */
	unsigned long		num_gamma_devices = 0;
	struct net_device	*gamma_devices[MAX_NUM_NICS];
/* Local ethernet addresses, initialized by gamma_start(). 
 * Used by make_send_header() macro to fill the sender field in eth frames header. */
	eth_addr		local_hw_addr[MAX_NUM_NICS];

#ifdef MORE_NICS_SAMENODE_ROUTING
/* 
 * Global arrays for adapters and rx_rings.
 * Initialized by gamma_start(). See under bind2/ for details. 
 */
	bind2_global_decl;
#endif

	task_t			*gamma_run_handler;
	task_t			*gamma_run_handler_samenode;
	unsigned long		gamma_irq_disabled;
	unsigned long		total_num_nodes = 0;
	unsigned long		phys_whoami;

	struct {
	  task_t   *task;
	  time_586  time;
	}			gamma_must_yield,
				gamma_must_yield_samenode;

/*  sys_gamma_suspend_irq()/sys_gamma_resume_irq() balance, and mutex  */
	spinlock_t		gamma_susp_irq_count_lock = SPIN_LOCK_UNLOCKED;
	unsigned long		gamma_susp_irq_count;

/*
 *  Table of (ptrs to) active parallel applications (one entry for each).
 *  Initialized by gamma_start(); entries allocated by sys_gamma_vm_create()
 *  and gamma_kernel_rec().
 */
	par_entry		*par_proc_table[MAX_PAR_PID+1];

/*
 *  Credit-based flow ctl machinery, with congestion control.
 *  Initialized by sys_gamma_getconfig() and gamma_start().
 */
	atomic_t		local_credit[MAX_NUM_NODES];
	unsigned long		initial_credit[MAX_NUM_NODES];
	unsigned long		credit_ceil[MAX_NUM_NODES];
	unsigned long		credit_growth_cnt[MAX_NUM_NODES];
	unsigned long		credit_size;

/*
 *  Per-destination rings of (ptrs to) per-packet tx temp buffers.
 *  Bcast pkts use entry #phys_whoami ("self", which would otherwise be
 *  unused).
 *  Alloc and init by gamma_start(), init by gamma_getconfig().
 */

typedef struct {
	void *b;		/*  Ptr to temp buffer  */
	long l;			/*
				 *  pkt from blocking send routines:
				 *    pkt size (>=0);
				 *  pkt from non-blocking send routines:
				 *    1 plus the NIC Tx ring tail at the time
				 *    when the pkt was enqueued,
				 *    then change sign (<0 therefore).
				 */
	unsigned long hlen;	/*
				 *  non-blocking send only:
				 *  size of header (+ possible 1st chunk,
				 *  in case 2-way gather send was used).
				 */
	unsigned long plen;	/*
				 *  non-blocking send only:
				 *  size of of payload (2nd chunk only,
				 *  in case 2-way gather send was used).
				 */
} tx_tmpbuf_t;

typedef struct {
	tx_tmpbuf_t buf[GAMMA_TX_WIN_SIZE];
	volatile u16 tail;
#ifdef SEQCHECK
/*
 *  Pkt numbering.  Each pkt is stamped by a sequence number.
 *  The current pkt number for destination node #N is stored here;
 *  the expected pkt number from source node #N is stored at rx_seq[N].
 */
	volatile u8 tx_seq;
/*  Pad to doubleword.  */
	u8 dummy;
#else
	u16 dummy;
#endif
} tx_tmp_ring_t;

	tx_tmp_ring_t		tx_tmp_ring[MAX_NUM_NODES];
	tx_tmp_ring_t		kernel_tx_tmp_ring;

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Ring of (ptrs to) temp buffers for intra-node message exchange.
 *  Init and alloc by gamma_start().
 */

#define IS_1ST_CNK 0x1
#define IS_LAST_CNK 0x2
#define IS_EMPTY 0x4
#define IN_USE 0x8

typedef struct {
	unsigned long par_pid;
	unsigned long procnum;
	unsigned long port;
	unsigned long cnk_len;
	unsigned long bitmap;
	unsigned long num_locdest;
	time_586 watch_time;
	task_t *sender;
	void *buf;
} shared_ring_entry;

typedef struct {
	volatile unsigned long head;
	volatile unsigned long dirty;
	volatile unsigned long tail;
	shared_ring_entry desc[SHARED_RING_SIZE];
} shared_ring_t;

	shared_ring_t		shared_ring;

#endif

#ifdef SEQCHECK
/*
 *  Table of pkt numbering.  Each pkt is stamped by a sequence number.
 *  The current pkt number for destination node #N is stored at tx_seq[N];
 *  the expected pkt number from source node #N is stored at rx_seq[N].
 *  Initialized by gamma_getconfig().
 *  Bcast pkts have distinct numbering.
 *  DO NOT CHANGE the type, unless also changing in the GAMMA header.
 */
	u8			rx_seq[MAX_NUM_NODES];
	u8			rx_seq_bcast[MAX_NUM_NODES];

/*
 *  Per-node table of flags: 1 iff pkts from given node must be rejected;
 *  1 iff given node has experienced pkt losses (congestion?).
 *  Initialized by sys_gamma_getconfig()
 */
	unsigned long		rx_discard[MAX_NUM_NODES];
	unsigned long		congested[MAX_NUM_NODES];
#endif

/*
 *  Counter of invocations to sys_gamma_poll(), plus a threshold.
 *  The counter is incremented each time a polling phase ends without getting
 *  any new pkt.
 *  When its value get higher than the threshold, the waiting node
 *  may suppose there is a dealock due to one or more missing pkts;
 *  it then asks everybody for their current frame number (by sending TEST
 *  pkt).
 *  This helps detecting missing pkts and thus keeps the GAMMA protocol alive.
 */

#define REPEAT_POLL 200                  /*  200 usec poll phase duration */
#define MAX_POLLS (1000000/REPEAT_POLL)  /*  1 sec timeout                */
#define CEILING_MAX_POLLS 1*MAX_POLLS    /*  Can be increased             */

	int			how_many_polls = 0;
	int			max_polls = MAX_POLLS;
	int			base_max_polls = MAX_POLLS;

/*  spin/lock semaphores for mutual excl on Tx/Rx  */
	spinlock_t		gamma_rx_lock = SPIN_LOCK_UNLOCKED;
	spinlock_t		gamma_tx_lock = SPIN_LOCK_UNLOCKED;
	spinlock_t		gamma_rxsamenode_lock = SPIN_LOCK_UNLOCKED;

/*
 *  Order of magnitude of sizeof(par_entry), needed for get_free_pages().
 *  Initialized by gamma_start();
 */
	unsigned long		par_entry_magnitude;

/*  Routing stuff  */
	eth_addr		eth_table[MAX_NUM_NODES];
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned short		routing_table[MAX_NUM_NODES];
#endif

	int			gamma_initialized = 0;

#ifdef MEASURE_OUT_TH
unsigned long out_counter = 0;
unsigned long out_total_amount = 0;
#endif

#ifdef MEASURE_OUT_TH_MEM
unsigned long out_counter_mem = 0;
unsigned long out_total_amount_mem_delay = 0;
unsigned long out_total_amount_mem_bytes = 0;
#endif

#ifdef MEASURE_IN_TH
unsigned long in_counter = 0;
unsigned long in_total_amount = 0;
#endif

#ifdef MEASURE_IN_TH_MEM
unsigned long in_counter_mem = 0;
unsigned long in_total_amount_mem_delay = 0;
unsigned long in_total_amount_mem_bytes = 0;
#endif


#ifdef MODULE
/************************************************************************
 *	Module support (experimental)					*
 ************************************************************************/

MODULE_LICENSE("GPL");
char *g_idt_table = NULL;
struct desc_struct *gidt_table;
module_param(g_idt_table,charp,0);
#endif


/************************************************************************
 *	Some function protos						*
 ************************************************************************/


long sys_gamma_poll(void);
static inline int is_low_watermark(unsigned long,unsigned long,unsigned long);


/************************************************************************
 *	Init routine							*
 ************************************************************************/


void gamma_start
(void)
{

	int				i,j;


#ifdef MODULE
	gidt_table = (struct desc_struct *)g_idt_table;
#endif

/*
 *  Allocate per-packet tx tmp bufs (one for each Tx ring entry).
 *  We don't yet know how many nodes are actually present in the cluster;
 *  in doubt, we must allocate as many Tx rings as MAX_NUM_NODES.
 *  Later, we de-allocate those one unneeded (gamma_getconfig()).
 */
	for (i=0; i < MAX_NUM_NODES; i++) {
		for (j=0; j < GAMMA_TX_WIN_SIZE; j++) {
			tx_tmp_ring[i].buf[j].b =
				kmalloc(FRAME_SIZ, GFP_KERNEL);
			if (tx_tmp_ring[i].buf[j].b == NULL) {
				printg("gamma_start():  kmalloc() failed\n");
				return;
			}
			tx_tmp_ring[i].buf[j].l = 0;
		}
		tx_tmp_ring[i].tail = 0;
#ifdef SEQCHECK
		tx_tmp_ring[i].tx_seq = 0;
#endif
	}
	for (j=0; j < GAMMA_TX_WIN_SIZE; j++) {
		kernel_tx_tmp_ring.buf[j].b =
			kmalloc(FRAME_SIZ, GFP_KERNEL);
		if (kernel_tx_tmp_ring.buf[j].b == NULL) {
			printg("gamma_start():  kmalloc() failed\n");
			return;
		}
		kernel_tx_tmp_ring.buf[j].l = 0;
	}
	kernel_tx_tmp_ring.tail = 0;
#ifdef SEQCHECK
	kernel_tx_tmp_ring.tx_seq = 0;
#endif
	gamma_run_handler = NULL;
	gamma_run_handler_samenode = NULL;
	gamma_must_yield.task = NULL;
	gamma_must_yield.time.hi = 0;
	gamma_must_yield.time.lo = 0;
	gamma_must_yield_samenode.task = NULL;
	gamma_must_yield_samenode.time.hi = 0;
	gamma_must_yield_samenode.time.lo = 0;
	gamma_irq_disabled = 0;
	gamma_susp_irq_count = 0;

/*  The table of parallel processes is empty initially  */
	for (i=1; i <= MAX_PAR_PID; i++)
		par_proc_table[i] = NULL;

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Init and allocate the ring of temp buffers
 *  for intra-node communication.
 */
	shared_ring.head = shared_ring.dirty = shared_ring.tail = 0;
	for (i=0; i < SHARED_RING_SIZE; i++) {
		shared_ring.desc[i].par_pid = 0;
		shared_ring.desc[i].procnum = 0;
		shared_ring.desc[i].port = 0;
		shared_ring.desc[i].cnk_len = 0;
		shared_ring.desc[i].bitmap = IS_EMPTY;
		shared_ring.desc[i].num_locdest = 0;
		shared_ring.desc[i].watch_time.hi = 0;
		shared_ring.desc[i].watch_time.lo = 0;
		shared_ring.desc[i].buf = kmalloc(PAGE_SIZE,GFP_KERNEL);
		if (shared_ring.desc[i].buf == NULL) {
			printg("gamma_start():  kmalloc() failed\n");
			return;
		}
		shared_ring.desc[i].sender = NULL;
	}
#endif

/*  Initial credit size is max (as defined by constant MAX_CREDIT_SIZE).  */
	credit_size = MAX_CREDIT_SIZE;

/*  Compute the magnitude of sizeof(par_entry), needed to allocate GAMMA vm  */
	for (i=1, j=0; i*PAGE_SIZE<sizeof(par_entry); i<<=1, j++);
	par_entry_magnitude = j;

/*  Set the entry points for the GAMMA primitives (trap 0x21)  */
	gamma_set_system_gate(0x21, &gamma_system_call);

#ifdef MORE_NICS_SAMENODE_ROUTING
/*  Initialize global arrays of adapters and rx_rings */
	bind2_global_init;
#endif

/* Initialize array of local mac addresses from probed nics. 
 * If multiple nics support is disabled, the first array entry 
 * is the only present nic. */
	for (i=0; i<num_gamma_devices; i++) {
		for (j=0; j<6; j++)
			local_hw_addr[i].addr[j] = gamma_devices[i]->dev_addr[j];
		local_hw_addr[i].addrl =
			((unsigned long)local_hw_addr[i].addr[0]) +
			((unsigned long)local_hw_addr[i].addr[1] << 8) +
			((unsigned long)local_hw_addr[i].addr[2] << 16) +
			((unsigned long)local_hw_addr[i].addr[3] << 24);
		local_hw_addr[i].addrw =
			((unsigned long)local_hw_addr[i].addr[4]) +
			((unsigned long)local_hw_addr[i].addr[5] << 8);
	}


	printg("\nGAMMA  Genoa Active Message MAchine (Linux 2.6)  "
		"1 Mar 2007\n");

#ifdef DRIVER_DEBUG_EVERYTHING
	printg("       Debug level: Very High\n");
#elif defined DRIVER_DEBUG
	printg("       Debug level: High\n");
#elif defined DRIVER_DEBUG_LIGHT
	printg("       Debug level: Low\n");
#else
	printg("       Debug level: Null\n");
#endif

#ifdef GAMMA_FULL_DUPLEX
	printg("       Full-duplex mode\n");
#else
	printg("       Half-duplex mode\n");
#endif
#ifdef SAFE_TX
	printg("       safe tx enabled\n");
#endif
#ifdef USE_JUMBO_FRAMES
	printg("       jumbo frame support enabled:  MTU %d\n",FRAME_SIZ);
#else
	printg("       jumbo frame support disabled!!\n");
#endif
#ifdef MORE_NICS_SAMENODE_ROUTING
	printg("       Multiple NICs support enabled\n");
        printg("       Found %ld devices:\n", num_gamma_devices);
	for(i=0; i < num_gamma_devices; i++) {
		printg("              Device %d: MAC Address %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n", i,
		       gamma_devices[i]->dev_addr[0],
		       gamma_devices[i]->dev_addr[1],
		       gamma_devices[i]->dev_addr[2],
		       gamma_devices[i]->dev_addr[3],
		       gamma_devices[i]->dev_addr[4],
		       gamma_devices[i]->dev_addr[5]); 
	}
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	printg("       multiple process instances per node: enabled\n");
	printg("               (max %d instances/node)\n", MAX_NUM_LOCAL_INST);
#else
	printg("       multiple process instances per node: disabled\n");
#endif
#ifdef NIC_IS_TG3
	printg("       Size of Tx ring %d\n", GAMMA_TX_RING_SIZE);
#else
	printg("       Size of Tx ring %d     Size of Rx ring %d\n",
		GAMMA_TX_RING_SIZE, GAMMA_RX_RING_SIZE);
#endif
	printg("       GAMMA protocol: %s\n",gamma_authors);
	printg("       BIND2 layer   : %s\n\n",bind2_authors);

	return;
}


/************************************************************************
 *	Service send/receive routines					*
 ************************************************************************/


/*  CAUTION:  do not use flow ctl here!  */
void gamma_kernel_send
(
	unsigned long phys_dest,
	unsigned char par_pid,
	unsigned char kind,
	void *data,
	unsigned short len,
	int use_lock
)
{

	tx_tmp_ring_t			*win;
	unsigned long			entry;
	l2s				*h;
	int				i;
	tx_common_decl;

	if (use_lock) spin_lock(&gamma_tx_lock);

#ifdef MORE_NICS_SAMENODE_ROUTING
	if (phys_dest != BROADCAST)
		tx_common_init( routing_table[phys_dest] );
	else {
		/* gamma_kernel_send broadcast is actually a software implementation. 
		 * should be changed in next releases. */
DEBUG_E(printg("gamma_kernel_send: entering software broadcast.\n"));

		for (i=0; i<total_num_nodes; i++) {
			if (i != phys_whoami)
				gamma_kernel_send(i,par_pid,kind,data,len,0);
		}
		if (use_lock) spin_unlock(&gamma_tx_lock);
		return;
	}
#else
	tx_common_init(0);
#endif

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG_E(printg("gamma_kernel_send: entering...\n"));
DEBUG_E(printg("     Send to physical node %lu through NIC %d\n",
	       phys_dest,
	       routing_table[phys_dest]));
#endif

/*
 *  Pkts of the flow-ctl protocol must be kept into the
 *  Tx tmp ring (for possible retransmission, and to keep the numbering
 *  consistent); other kernel pkts need not.
 */
	if (kind == FILL_CREDIT || kind == CRED_REQ)
		if (phys_dest != BROADCAST)
			win = &tx_tmp_ring[phys_dest];
		else
			win = &tx_tmp_ring[phys_whoami];
	else
		win = &kernel_tx_tmp_ring;

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;

	if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:
#ifdef MORE_NICS_SAMENODE_ROUTING
	/* Now broadcast is a sw one, but this code is ready for hw implementation. */
	make_send_header(phys_dest,par_pid,kind,0,ETH_GAMMA_KERNEL,1,len,0,h,
			 (phys_dest == BROADCAST)? 0 : routing_table[phys_dest]);
#else
	make_send_header(phys_dest,par_pid,kind,0,ETH_GAMMA_KERNEL,1,len,0,h,0);
#endif

#ifdef SEQCHECK
	if (kind == FILL_CREDIT || kind == CRED_REQ) {
/*
 *  Pkts of the flow-ctl protocol must be kept numbered,
 *  as they are kept into the Tx tmp ring for possible retransmission.
 */
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
	}

DEBUG_F(
if (kind == FILL_CREDIT)
	printg("send FILL_CREDIT to node %ld, frame_no %d\n",
		phys_dest,read_send_header_frameno(h));
if (kind == CRED_REQ)
	printg("send CRED_REQ to node %ld, frame_no %d\n",
		phys_dest,read_send_header_frameno(h));
if (kind == ACK)
	printg("send ACK to node %ld, frame_no %d\n",
		phys_dest,*(unsigned char *)data);
if (kind == ACKB)
	printg("send ACKB to node %ld, frame_no %d\n",
		phys_dest,*(unsigned char *)data);
)

#endif
	memcpy((u8 *)h+HEADER_SIZE,data,len);

DEBUG_E(printg(
"gamma_kernel_send(): %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x\n",
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll, h[5].ll
));

DEBUG(printg(
"gamma_kernel_send(): sending to node %ld, kind %d\n",
phys_dest,kind
));

	tx_enq_pkt1(h,HEADER_SIZE+len);
	tx_start();
	tx_forw_tail();

/*
 * this is a workaround to a hw bug of e1000: the tx engine occasionally hangs.
 * Flushing the tx ring can detect and solve this in a few seconds (!!)
 */
	if (kind == ACK || kind == TEST)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);

#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	win->buf[entry].l = HEADER_SIZE+len;

end:

	if (use_lock) spin_unlock(&gamma_tx_lock);
	return;

ring_full:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (i) goto room_in_tx_ring;

DEBUG_L(printg("gamma_kernel_send(): Transmit failure before send.\n"));

	goto end;

}


#ifdef SEQCHECK
/*
 *  Scan a given tx tmp ring to retrieve pkts which are to be retransmitted.
 */
static int gamma_scan_tx_tmp_ring
(unsigned long phys_dest, u32 frameno, int is_bcast)
{

	tx_tmp_ring_t			*win;
	int				i;
	l2s				*h;


	if (!is_bcast)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

/*  scan the tx tmp ring backward (modulo ring size)  */
	for (i = win->tail-1; i != win->tail; i--) {

		if (i < 0) i += GAMMA_TX_WIN_SIZE;
		h = win->buf[i].b;

DEBUG(printg(
"gamma_scan_tx_tmp_ring(): (search %u, pos %d) %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x\n",
frameno, i, h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll, h[5].ll
));

		if (read_send_header_frameno(h) != frameno)
			continue;
/*  Pkt found.  */
		return i;

	}

/*  Pkt not found.  */
	return -1;

}


/*  Re-enqueue pkts to be retransmitted.  */
/*  CAUTION:  do not use flow ctl here!  */
static int gamma_retx
(unsigned long phys_dest, unsigned long pos, int is_bcast)
{

	tx_tmp_ring_t			*win;
	int				ii;
	int				tx_handle;
	l2s				*h;
	int				i;
	tx_common_decl;


#ifdef MORE_NICS_SAMENODE_ROUTING
	tx_common_init( routing_table[phys_dest] );
#else
	tx_common_init(0);
#endif

	if (!is_bcast)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

/*  loop over the tx tmp ring from "pos" to end.  */
	for (ii = pos; ; ) {

		h = win->buf[ii].b;
		tx_handle = -(win->buf[ii].l) - 1;
		if (tx_handle >= 0) {
#if 0
/*
 *  Pkt is from a non-blocking send.  We must check it is still pending in
 *  the Tx ring so that we can still retransmit it
 *  (according to the non-blocking semantics of isend() routines).
 */
			if ((!(tx_ring_occupancy))
			|| (tx_handle > tx_ring_tail)
			|| ((tx_ring_head < tx_ring_tail)
				&& (tx_handle <= tx_ring_head))
		) {
#endif

DEBUG_L(printg(
"Could not retransmit a flushed pkt originated from a non-blocking send()\n"
));

				goto skip;
#if 0
			}
#endif
		}
/*
 *  This pkt is OK for retransmission.  But we have to change its type
 *  to "retransmitted".
 *  Rationale:  pkts in the Tx tmp ring can be bcast; if we retransmit them,
 *  we must prevent non-involved nodes to catch them twice!
 */
		if (rec_header_frmtype(h) == ETH_GAMMA)
			set_send_header_frmtype(h,ETH_GAMMA_RETX);
		else if (rec_header_frmtype(h) == ETH_GAMMA_KERNEL)
			set_send_header_frmtype(h,ETH_GAMMA_KERNEL_RETX);
/*  Enqueue pkt for retransmit.  */
#if 0
		if (tx_handle >= 0) {
printg("ciao retx isend pkt, addr %8.8lx\n",
tx_entry_2_payloadptr((tx_handle+GAMMA_TX_RING_SIZE-1) % GAMMA_TX_RING_SIZE));
/*  Pkt from non-blocking send:  copy payload from user to kernel space.  */
			memcpy(
((u8 *)h) + win->buf[i].hlen,
tx_entry_2_payloadptr((tx_handle+GAMMA_TX_RING_SIZE-1) % GAMMA_TX_RING_SIZE),
win->buf[i].plen
			);
printg("ciao fatto\n");
		}
#endif
		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

DEBUG_F(printg(
"gamma_retx(): -> %ld, frm %d, pos %d: %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x\n",
phys_dest,read_send_header_frameno(h),ii,
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll, h[5].ll
));

		tx_enq_pkt1(h,win->buf[ii].l);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif

skip:

		incr_modulo(ii,GAMMA_TX_WIN_SIZE);
		if (ii == win->tail)
			break;

	}

	return 0;

ring_full:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (i) goto room_in_tx_ring;

DEBUG_L(printg("gamma_retx(): Transmit failure before send.\n"));

	return -1;

}
#endif


static void gamma_kernel_rec
(
	void *addr,
	unsigned char par_pid,
	unsigned short len,
	unsigned char kind,
	unsigned long phys_sender
)
{

	par_entry			*p;
	port_entry			*pp;
	u32				frame_no;
	int				i,j,pos,procnum;


	p = par_proc_table[par_pid];

	switch (kind) {

	case FILL_CREDIT:

#ifdef SEQCHECK
		if (!congested[phys_sender]) {
			if (initial_credit[phys_sender]
				>= credit_ceil[phys_sender]
			) {
				/*  Credit size raises the allowed ceiling.  */
				if (credit_ceil[phys_sender] >= credit_size)
					/*  Credit ceiling is max.  */
					goto out;
				/*
				 *  Credit ceiling can grow.
				 *  We gradually (and slowly) push up the
				 *  ceiling until max is reached.
				 */
				if (++credit_growth_cnt[phys_sender] >= 1000) {
					credit_growth_cnt[phys_sender] = 0;
					credit_ceil[phys_sender]++;
				}
				goto out;
			}
			/*
			 *  Push up the credit size till current ceiling.
			 *  Do not push up the ceiling yet.
			 */
			initial_credit[phys_sender]++;
			goto out;
		}
/*  A possible congestion on node "phys_sender".  Slow down...  */
		congested[phys_sender] = 0;
/*  Shrink down the credit ceiling (just a bit)  */
		credit_ceil[phys_sender]
			= gmax(2,initial_credit[phys_sender]-1);
		credit_growth_cnt[phys_sender] = 0;
/*  Shrink down the credit size (quite a lot)  */
		if (initial_credit[phys_sender] > 2)
			initial_credit[phys_sender]
				= gmax(2,initial_credit[phys_sender]>>1);
/*  This is a less aggressive shrink rule...I could not do any meaningful
comparison with the previous one yet :(
				= gmax(2,initial_credit[phys_sender]-1);
*/
out:

#endif
		atomic_add(
			initial_credit[phys_sender],
			&local_credit[phys_sender]);
		return;

	case CRED_REQ:

	 	gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
		return;

#ifdef SEQCHECK
	case ACK:

/*  Check the frame numbering of "phys_sender" (unicast).  */
		frame_no = *(u32 *)addr;
		if (frame_no == rx_seq[phys_sender]) {

DEBUG(printg("ACK from node %ld, frame_no %u\n",phys_sender,frame_no))

/*  No missing pkt was detected. Increase the time interval to next check.  */
			if (max_polls < CEILING_MAX_POLLS)
				max_polls +=
				base_max_polls/(2*(total_num_nodes-1));
			return;
		}
/*  Missing pkts from "phys_sender". See rx_gamma(), label "err_frame_no"  */

DEBUG_L(printg(
"gamma_kernel_rec(): ACK #%u, expected #%u, phys sender %ld\n",
frame_no,rx_seq[phys_sender],phys_sender
))

		if (((frame_no-rx_seq[phys_sender]+256)%256)
			> MAX_CREDIT_SIZE + LOW_WATERMARK) {

DEBUG_L(printg("Unfeasible ack?  Discard...\n"));

			return;
		}

		max_polls = base_max_polls;
		gamma_kernel_send(BROADCAST,DONT_MIND,CONGESTION,NULL,0,1);
		frame_no = rx_seq[phys_sender];
		gamma_kernel_send(
			phys_sender,DONT_MIND,NACK,&frame_no,
			sizeof(frame_no),1);
		rx_discard[phys_sender] = 1;
		return;

	case ACKB:

/*  Check the frame numbering of "phys_sender" (bcast).  */

		frame_no = *(u32 *)addr;
		if (frame_no == rx_seq_bcast[phys_sender]) {

DEBUG(printg("ACKB from node %ld, frame_no %u\n",phys_sender,frame_no))

/*  No missing pkt was detected. Increase the time interval to next check.  */
			if (max_polls < CEILING_MAX_POLLS)
				max_polls +=
				base_max_polls/(2*(total_num_nodes-1));
			return;
		}
/*  Missing pkts from "phys_sender". See rx_gamma(), label "err_frame_no"  */

DEBUG_L(printg(
"gamma_kernel_rec(): ACKB #%u, expected #%u, phys sender %ld\n",
frame_no,rx_seq_bcast[phys_sender],phys_sender
))

		if (((frame_no-rx_seq_bcast[phys_sender]+256)%256)
			> MAX_CREDIT_SIZE + LOW_WATERMARK) {

DEBUG_L(printg("Unfeasible bcast ack?  Discard...\n"));

			return;
		}

		frame_no = rx_seq_bcast[phys_sender];
		gamma_kernel_send(
			phys_sender,DONT_MIND,NACKB,&frame_no,
			sizeof(frame_no),1);
		rx_discard[phys_sender] = 1;
		return;

	case CONGESTION:

/*
 *  A possible congestion on node "phys_sender".  Slow down at next credit
 *  renewal.
 */
		congested[phys_sender] = 1;
		return;

	case NACK:

/*  NACK for missing unicast pkts.  */

                frame_no = *(u32 *)addr;

DEBUG_L(printg("NACK from node %ld, frame %u\n",phys_sender,frame_no));

		spin_lock(&gamma_tx_lock);
		pos = gamma_scan_tx_tmp_ring(phys_sender,frame_no,0);
		if (pos >= 0) {
/*  Go-back-N and retransmit.  */
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_START,NULL,0,0);
			gamma_retx(phys_sender,pos,0);
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_END,NULL,0,0);
			spin_unlock(&gamma_tx_lock);
			return;
		}

DEBUG_L(printg(
"Missing pkt %u from node %ld:  not found. Cannot retransmit!\n",
frame_no,phys_sender
));

		gamma_kernel_send(phys_sender,DONT_MIND,ERR_NO_RETX,addr,len,0);
		spin_unlock(&gamma_tx_lock);
		return;

	case NACKB:

/*  NACK for missing bcast pkts.  */

                frame_no = *(u32 *)addr;

DEBUG_L(printg("NACKB from node %ld, bcast frame %u\n",phys_sender,frame_no));

		spin_lock(&gamma_tx_lock);
		pos = gamma_scan_tx_tmp_ring(phys_sender,frame_no,1);
		if (pos >= 0) {
/*  Go-back-N and retransmit.  */
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_START,NULL,0,0);
			gamma_retx(phys_sender,pos,1);
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_END,NULL,0,0);
			spin_unlock(&gamma_tx_lock);
			return;
		}

DEBUG_L(printg(
"Missing bcast pkt %u from node %ld:  not found. Cannot retransmit!\n",
frame_no,phys_sender
));

		gamma_kernel_send(phys_sender,DONT_MIND,ERR_NO_RETX,addr,len,0);
		spin_unlock(&gamma_tx_lock);
		return;

	case TEST:

/*
 *  This node is asked to spread its table of frame numbering (unicast
 *  as well as bcast);
 *  this can help recovering from a deadlock in case of pkt losses.
 */

DEBUG(printg("TEST from node %ld\n",phys_sender))

		if (total_num_nodes == 0)
/*
 *  Maybe this machine has rebooted and has not yet called
 *  sys_gamma_getconfig(), while other ones are waiting for pkts.
 *  Discard this request.
 */
			return;

		spin_lock(&gamma_tx_lock);

		frame_no = tx_tmp_ring[phys_sender].tx_seq;
		gamma_kernel_send(
			phys_sender,DONT_MIND,ACK,
			&frame_no,sizeof(frame_no),0);
		frame_no = tx_tmp_ring[phys_whoami].tx_seq;
		gamma_kernel_send(
			phys_sender,DONT_MIND,ACKB,
			&frame_no,sizeof(frame_no),0);

		spin_unlock(&gamma_tx_lock);

		return;

	case RETX_START:

/*
 *  A sequence of retransmitted pkts begins.
 *  Since this may occur only after detecting a missing pkt in the
 *  sequence from node "phys_sender", we should have set
 *  rx_discard[phys_sender] = 1.
 */
		if (rx_discard[phys_sender] == 1) {

DEBUG_F(printg("Start of retx sequence, sender %ld\n",phys_sender));

			rx_discard[phys_sender] = 2;
			return;
		}

DEBUG_F(printg(
"gamma_kernel_rec():  Start of double retransmission from node %ld(!!!)\n",
phys_sender
));

		return;

	case RETX_END:

/*
 *  A sequence of retransmitted pkts ends.
 *  Since this may occur only after a sequence of retransmitted pkts
 *  from node "phys_sender", we should have set
 *  rx_discard[phys_sender] = 2.
 */
		if (rx_discard[phys_sender] == 2) {

DEBUG_F(printg("End of retx sequence, sender %ld\n",phys_sender));

			rx_discard[phys_sender] = 0;
			return;
		}
		if (rx_discard[phys_sender] == 1) {
/*
 *  Asked for retransmission but did not catch the RETX_START pkt;
 *  or, detected missing pkt during a retransmission sequence and
 *  therefore refusing remaining part of sequence, but did not yet get
 *  RETX_START.
 */

DEBUG_F(printg(
"gamma_kernel_rec():  unexpected RETX_END (missing RETX_START?)\n"
));

			return;
		}

DEBUG_F(printg(
"gamma_kernel_rec():  End of double retransmission from node %ld(!!!)\n",
phys_sender
));

		return;
#endif  /*  SEQCHECK  */

	case VM_WISH:

/*  A remote proc wishes to join virtual GAMMA #par_pid  */

		if (p != NULL) {
/*  Virtual GAMMA already created.  If not a stale express the wish  */
			for (i=0; i < MAX_NUM_LOCAL_INST; i++) {
				if (p->inst_table[i].pid != -1 ||
				p->inst_table[i].task != NULL ||
				p->inst_table[i].whoami != -1) {
printg("gamma_kernel_rec(): failed to join virtual GAMMA #%d: stale exists\n",
par_pid);
					return;
				}
			}
			p->vm.num_instances++;
			p->wish_flag[phys_sender]++;
			return;
		}
/*  Allocate a descriptor for a new virtual GAMMA  */
		p = (par_entry *)__get_free_pages(GFP_ATOMIC,
						par_entry_magnitude);
		if (p == NULL) {
printg("gamma_kernel_rec(): kmalloc() failed creating virtual GAMMA #%d\n",
par_pid);
			return;
		}
		for (i=0; i < MAX_NUM_NODES; i++)
			p->wish_flag[i] = 0;
		for (i=0; i < MAX_NUM_LOCAL_INST; i++) {
			p->inst_table[i].pid = -1;
			p->inst_table[i].task = NULL;
			p->inst_table[i].whoami = -1;
		}
/*  Express the wish, then return.  */
		p->vm.num_instances = 1;
		p->wish_flag[phys_sender]++;
		par_proc_table[par_pid] = p;
		return;

	case VM_ABORT:

		if (p == NULL) {

DEBUG(printg(
"gamma_kernel_rec(): trying to abort non-existant virtual GAMMA #%d from remote node\n",
par_pid
));

			return;
		}

		if (p->vm.random_key != *(typeof(p->vm.random_key) *)addr) {

DEBUG(printg(
"gamma_kernel_rec(): trying to abort wrong virtual GAMMA #%d from remote node\n",
par_pid
));

			return;
		}

		if (p->quitting) {

DEBUG(printg(
"gamma_kernel_rec(): virtual GAMMA #%d already quitting\n",par_pid
));

			return;
		}

		p->quitting = 1;

/*  Ask all local instances (if any) to abort (SIGTERM sighandler)  */
		for (i=0; i < MAX_NUM_LOCAL_INST; i++)
			if (
				p->inst_table[i].whoami >= 0 &&
				p->inst_table[i].pid > 0 &&
				p->inst_table[i].task != NULL &&
				p->inst_table[i].task->state != EXIT_ZOMBIE &&
				p->inst_table[i].task->state != EXIT_DEAD
			) {
				force_sig(SIGTERM, p->inst_table[i].task);

DEBUG_L(printg("gamma_kernel_rec(): aborting par_pid %d procnum %d...\n",
par_pid,i));

			}

		return;


	case VM_CREATE:

/*
 *  It is possible that an entry for virtual GAMMA with id #par_pid
 *  be already allocated due to a previous call to sys_gamma_vm_wish()
 *  (on a local as well as remote node).  So, no error if entry already
 *  allocated.
 */
		if (p == NULL) {
/*  Allocate a descriptor for a new virtual GAMMA  */
			p = (par_entry *)__get_free_pages(GFP_ATOMIC,
						par_entry_magnitude);
			if (p == NULL) {
printg("gamma_kernel_rec(): kmalloc() failed creating virtual GAMMA #%d\n",
par_pid);
				return;
			}
			memset(p,0,sizeof(par_entry));
			for (i=0; i < MAX_NUM_LOCAL_INST; i++) {
				p->inst_table[i].pid = -1;
				p->inst_table[i].task = NULL;
				p->inst_table[i].whoami = -1;
			}
		} else {
/* fail/ignore if this is a stale virtual GAMMA */
			for (i=0; i < MAX_NUM_LOCAL_INST; i++) {
				if (p->inst_table[i].pid != -1 ||
				p->inst_table[i].task != NULL ||
				p->inst_table[i].whoami != -1) {
#if 0
printg("gamma_kernel_rec(): failed to create virtual GAMMA #%d: stale exists\n",
par_pid);
#endif
					return;
				}
			}
		}

/*  Get and store the virtual GAMMA descriptor  */
                memcpy(&(p->vm),addr,len);
/*
 *  Compute the table of ethernet addresses of the new virtual GAMMA, and
 *  instance # ("whoami") of all the "procnum" processes to be run on the
 *  local node
 *  (-1 iff local node is not involved at all in the virtual GAMMA).
 */
		procnum = 0;
		for (i=0; i < p->vm.num_instances; i++) {
			p->eth_table[i] = eth_table[p->vm.node_list[i]];
			if (p->vm.node_list[i] == phys_whoami) {
/*
 *  Involved node: set "whoami" to the instance # and allocate room for
 *  per-port pre-computed pkt headers
 */
				procnum++;
				if (procnum > MAX_NUM_LOCAL_INST) {
printg("gamma_kernel_rec():  too many proc instances on this node when creating virtual GAMMA %d.  Aborting...\n",par_pid);
/*  Abort virtual GAMMA locally and globally.  */
					__sys_gamma_vm_abort(par_pid,1,0);
					return;
				}
				p->inst_table[procnum-1].whoami = i;
				for (j=0; j <= MAX_PORT_NUMBER; j++) {
					pp = &p->inst_table[procnum-1]
							.port_table[j];
					pp->semaphore =
					pp->boundary_1p =
					pp->in_size =
					pp->in_room = 0;
					pp->in_buffer =
					pp->in_progress = NULL;
					pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
					pp->partner_is_same_node =
					pp->bcast_was_susp = 0;
#endif
					pp->header = kmalloc(
						HEADER_SIZE+MAX_1ST_PIECE_SIZE,
						GFP_ATOMIC);
					if (pp->header == NULL) {
printg("gamma_kernel_rec(): kmalloc() failed when creating virtual GAMMA #%d\n",par_pid);
						return;
					}
				}
			}
		}

		p->num_locinst = procnum;
		par_proc_table[par_pid] = p;

DEBUG(
{
int k;
if (procnum == 1) {
	printg("gamma_kernel_rec():  Virtual GAMMA %d: num_instances %d, \n",
		par_pid, p->vm.num_instances);
	printg("node list: \n");
	for (k=0; k<p->vm.num_instances; k++)
		printg("%d - ",p->vm.node_list[k]);
	printg("--\n");
	printg("phys_whoami %ld\n",phys_whoami);
	for (k=0; k<p->vm.num_instances; k++)
		printg(
			"%d: %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
			k,
			p->eth_table[k].addr[0],
			p->eth_table[k].addr[1],
			p->eth_table[k].addr[2],
			p->eth_table[k].addr[3],
			p->eth_table[k].addr[4],
			p->eth_table[k].addr[5]
		);
}
if (procnum > 0)
	for (k=1; k<=procnum; k++)
		printg("Local instance %d\n",p->inst_table[k-1].whoami);
}
);

		return;

	} /* switch */

	return;

}


#ifdef MORE_INSTANCES_ON_SAME_NODE
/************************************************************************
 *	Intra-node send/receive routines				*
 ************************************************************************/


static int gamma_send_2p_samenode
(
	unsigned long my_par_pid,
	unsigned long my_procnum,
	unsigned long out_port,
	void *data1, unsigned long len1,
	void *data2, unsigned long len2,
	unsigned long is_bcast
)
{

	port_entry			*out_p, *dest_p;
	unsigned long			dest_par_pid;
	unsigned long			dest_procnum;
	unsigned long			dest_port;
	unsigned long			len;
	shared_ring_entry		*desc;
	unsigned long			cnk_len;
	task_t				*receiver;
	unsigned long			num_locdest = 1;
	unsigned long			i;


	out_p = &par_proc_table[my_par_pid]->inst_table[my_procnum]
						.port_table[out_port];
	dest_par_pid = out_p->partner_par_pid;
	dest_procnum = out_p->partner_procnum;
	dest_port = out_p->partner_port;
	dest_p = &par_proc_table[dest_par_pid]->inst_table[dest_procnum]
						.port_table[dest_port];

	if (is_bcast) goto send_bcast;

DEBUG(printg("gamma_send_2p_samenode(): enter: len1 %ld, len2 %ld\n",
len1,len2));

send_bcast_back:

	receiver = dest_p->task;
	if (receiver == NULL) goto no_partner;

	len = len1+len2;

DEBUG(printg(
"gamma_send_2p_samenode(): my_procnum %ld, num_locdest %ld, dest_procnum %ld\n",
my_procnum,num_locdest,dest_procnum
));

	if (shared_ring_room(SHARED_RING_SIZE) <= 4) goto ring_full;

	desc = &shared_ring.desc[shared_ring.dirty]; /* the next avail desc */
	incr_modulo(shared_ring.dirty,SHARED_RING_SIZE);

	desc->par_pid = dest_par_pid;
	desc->procnum = dest_procnum;
	desc->port = dest_port;
	cnk_len = gmin(len,PAGE_SIZE);
	desc->cnk_len = cnk_len;
	if (cnk_len < len)
		desc->bitmap = IS_1ST_CNK;
	else
		desc->bitmap = IS_1ST_CNK | IS_LAST_CNK;
	desc->num_locdest = num_locdest;
	desc->watch_time.hi = 0;
	desc->watch_time.lo = 0;
	desc->sender = current;
	gamma_copy_from_user(desc->buf,data1,len1);

DEBUG(printg(
"gamma_send_2p_samenode(): par_pid %ld, procnum %ld, port %ld, len %ld\n",
desc->par_pid,desc->procnum,desc->port,len
));

	for (;;) {

DEBUG(printg(
"gamma_send_2p_samenode(): cnk_len %ld, head %ld, tail %ld, dirty %ld\n",
desc->cnk_len,shared_ring.head,shared_ring.tail,shared_ring.dirty
));

		gamma_copy_from_user(
			(u8 *)(desc->buf)+len1,
			data2,cnk_len-len1);

/*  commit this chunk  */
		desc->bitmap |= IN_USE;
		incr_modulo(shared_ring.tail,SHARED_RING_SIZE);

		len -= cnk_len;
		if (!len) break;

/*  Next chunk...  */
		data2 = (u8 *)data2+cnk_len-len1;
		len1 = 0;

		if (shared_ring_room(SHARED_RING_SIZE) <= 4) goto ring_full;

		desc = &shared_ring.desc[shared_ring.dirty];
		incr_modulo(shared_ring.dirty,SHARED_RING_SIZE);

		desc->par_pid = dest_par_pid;
		desc->procnum = dest_procnum;
		desc->port = dest_port;
		cnk_len = gmin(len,PAGE_SIZE);
		desc->cnk_len = cnk_len;
		if (cnk_len < len)
			desc->bitmap = 0;
		else
			desc->bitmap = IS_LAST_CNK;
		desc->num_locdest = num_locdest;
		desc->watch_time.hi = 0;
		desc->watch_time.lo = 0;
		desc->sender = current;

	}

/*  Exit  */

DEBUG(printg(
"gamma_send_2p_samenode():  Done! head %ld, tail %ld\n",
shared_ring.head,shared_ring.tail
));

	return 0;

send_bcast:

/*  How many receivers on this node?  */
	if (out_p->partner_par_pid == my_par_pid)
/*  Sending to same group:  all members but sender.  */
		num_locdest = par_proc_table[my_par_pid]->num_locinst-1;
	else
/*  Sending to different group:  all members of dest group.  */
		num_locdest = par_proc_table[out_p->partner_par_pid]
								->num_locinst;
/*
 *  In case of bcast on local node, the receiver with smallest
 *  "procnum" is appointed to be the first to operate the receive
 *  (all entries produced in the shared ring are marked with its "procnum").
 *  A "propagation" mechanism is implemented in rx_gamma_samenode().
 *  Take note of the smallest "procnum" among receivers.
 */
	for (i=0; i < MAX_NUM_LOCAL_INST; i++) {
		if (par_proc_table[out_p->partner_par_pid]->inst_table[i]
			.whoami < 0
		)
			/*  No more local receivers  */
			break;
		if (out_p->partner_par_pid == my_par_pid) {
/*  Sending to same group:  sender will not receive.  */
			if (i != my_procnum) {
				dest_procnum = i;
				break;
			}
		} else {
			dest_procnum = i;
			break;
		}
	}
/*  Do we have local receivers actually?  */
	if (num_locdest) {

DEBUG(printg("gamma_send_2p_samenode(): enter: len1 %ld, len2 %ld\n",
len1,len2));
DEBUG(printg(
"gamma_send_2p_samenode(): bcast! my_procnum %ld, num_locdest %ld, dest_procnum %ld, len1 %ld, len2 %ld\n",
my_procnum,num_locdest,dest_procnum,len1,len2
));

		goto send_bcast_back;
	}
/*  No local receivers:  exit. */
	return 0;

ring_full:

DEBUG_L(printg(
"gamma_send_2p_samenode(): pid %d:  no flow ctl --> ring full.  Failure.\n",
current->pid));

no_partner:

	return -GAMMA_ERR_GENERIC;

}


static int gamma_send_2p_flowctl_samenode
(
	unsigned long my_par_pid,
	unsigned long my_procnum,
	unsigned long out_port,
	void *data1, unsigned long len1,
	void *data2, unsigned long len2,
	unsigned long is_bcast,
	unsigned long was_susp
)
{

	port_entry			*out_p, *dest_p;
	unsigned long			dest_par_pid;
	unsigned long			dest_procnum;
	unsigned long			dest_port;
	shared_ring_entry		*desc;
	unsigned long			len;
	unsigned long			cnk_len;
	task_t				*receiver;
	unsigned long			num_locdest = 1;
	unsigned long			i;


	out_p = &par_proc_table[my_par_pid]->inst_table[my_procnum]
						.port_table[out_port];
	dest_par_pid = out_p->partner_par_pid;
	dest_procnum = out_p->partner_procnum;
	dest_port = out_p->partner_port;
	dest_p = &par_proc_table[dest_par_pid]->inst_table[dest_procnum]
						.port_table[dest_port];

	if (is_bcast) goto send_bcast;

DEBUG(printg(
"gamma_send_2p_flowctl_samenode(): enter: len1 %ld, len2 %ld\n",len1,len2));

send_bcast_back:

	receiver = dest_p->task;
	if (receiver == NULL) goto no_partner;

	if (shared_ring_room(SHARED_RING_SIZE) <= SHARED_RING_LOW)
		goto ring_full;

	len = len1+len2;

DEBUG(printg(
"gamma_send_2p_flowctl_samenode(): my_procnum %ld, num_locdest %ld, dest_procnum %ld\n",
my_procnum,num_locdest,dest_procnum
));

	desc = &shared_ring.desc[shared_ring.dirty]; /* the next avail desc */
	incr_modulo(shared_ring.dirty,SHARED_RING_SIZE);

	desc->par_pid = dest_par_pid;
	desc->procnum = dest_procnum;
	desc->port = dest_port;
	cnk_len = gmin(len,PAGE_SIZE);
	desc->cnk_len = cnk_len;
	if (was_susp) {
		if (cnk_len < len)
			desc->bitmap = 0;
		else
			desc->bitmap = IS_LAST_CNK;
	} else if (cnk_len < len)
		desc->bitmap = IS_1ST_CNK;
	else
		desc->bitmap = IS_1ST_CNK | IS_LAST_CNK;
	desc->num_locdest = num_locdest;
	desc->watch_time.hi = 0;
	desc->watch_time.lo = 0;
	desc->sender = current;
	gamma_copy_from_user(desc->buf,data1,len1);

DEBUG(printg(
"gamma_send_2p_flowctl_samenode(): par_pid %ld, procnum %ld, port %ld, len %ld\n",
desc->par_pid,desc->procnum,desc->port,len
));

	for (;;) {

DEBUG(printg(
"gamma_send_2p_flowctl_samenode(): cnk_len %ld, head %ld, tail %ld, dirty %ld\n",
desc->cnk_len,shared_ring.head,shared_ring.tail,shared_ring.dirty
));

		gamma_copy_from_user(
			(u8 *)(desc->buf)+len1,
			data2,cnk_len-len1);

/*  commit this chunk  */
		desc->bitmap |= IN_USE;
		incr_modulo(shared_ring.tail,SHARED_RING_SIZE);

		len -= cnk_len;
		if (!len) break;

/*  Next chunk...  */
		data2 = (u8 *)data2+cnk_len-len1;
		len1 = 0;

		if (shared_ring_room(SHARED_RING_SIZE) <= 4) goto wait_room;

room_ok:

		desc = &shared_ring.desc[shared_ring.dirty];
		incr_modulo(shared_ring.dirty,SHARED_RING_SIZE);

		desc->par_pid = dest_par_pid;
		desc->procnum = dest_procnum;
		desc->port = dest_port;
		cnk_len = gmin(len,PAGE_SIZE);
		desc->cnk_len = cnk_len;
		if (cnk_len < len)
			desc->bitmap = 0;
		else
			desc->bitmap = IS_LAST_CNK;
		desc->num_locdest = num_locdest;
		desc->watch_time.hi = 0;
		desc->watch_time.lo = 0;
		desc->sender = current;
		continue;

wait_room:

		wait_room_in_shared_ring(i);
		if (i) goto room_ok;
/*  still no room after waiting time.  suspend...  */
		break;

	}

	if (!len) {
/*  Transmission complete.  Exit  */

DEBUG(printg(
"gamma_send_2p_flowctl_samenode():  Done! head %ld, tail %ld\n",
shared_ring.head,shared_ring.tail
));

		return 0;
	}
/*
 *  Transmission suspended due to ring full.
 */
	if (is_bcast) {
/*
 *  Take note that the operation was suspended while bcasting in the
 *  same node.
 */
		out_p->bcast_was_susp = 1;

DEBUG(printg("gamma_send_2p_flowctl_samenode(): suspend bcast.\n"));

	}

DEBUG(printg(
"gamma_send_2p_flowctl_samenode(): head %ld,tail %ld: No room.  Suspend (residual len %ld)\n",
shared_ring.head,shared_ring.tail,len
));

/*  Return how many bytes are still to be transmitted.  */
	return len;

send_bcast:

	if (out_p->bcast_was_susp) {
/*  Bcast in the same node had been suspended due to ring full.  */
		out_p->bcast_was_susp = 0;

DEBUG(printg("gamma_send_2p_flowctl_samenode(): resume bcast.\n"));

	}
/*  How many receivers on this node?  */
	if (out_p->partner_par_pid == my_par_pid)
/*  Sending to same group:  all members but sender.  */
		num_locdest = par_proc_table[my_par_pid]->num_locinst-1;
	else
/*  Sending to different group:  all members of dest group.  */
		num_locdest = par_proc_table[out_p->partner_par_pid]
								->num_locinst;
/*
 *  In case of bcast on local node, the receiver with smallest
 *  "procnum" is appointed to be the first to operate the receive
 *  (all entries produced in the shared ring are marked with its "procnum").
 *  A "propagation" mechanism is implemented in rx_gamma_samenode().
 *  Take note of the smallest "procnum" among receivers.
 */
	for (i=0; i < MAX_NUM_LOCAL_INST; i++) {
		if (par_proc_table[out_p->partner_par_pid]->inst_table[i]
			.whoami < 0
		)
			/*  No more local receivers  */
			break;
		if (out_p->partner_par_pid == my_par_pid) {
/*  Sending to same group:  sender does not count as receiver.  */
			if (i != my_procnum) {
				dest_procnum = i;
				break;
			}
		} else {
			dest_procnum = i;
			break;
		}
	}
/*  Do we have local receivers actually?  */
	if (num_locdest) {

DEBUG(printg(
"gamma_send_2p_flowctl_samenode(): enter: len1 %ld, len2 %ld\n",len1,len2
));
DEBUG(printg(
"gamma_send_2p_flowctl_samenode(): bcast! my_procnum %ld, num_locdest %ld, dest_procnum %ld, len1 %ld, len2 %ld\n",
my_procnum,num_locdest,dest_procnum,len1,len2
));

		goto send_bcast_back;
	}
/*  No local receivers:  exit. */
	return 0;

ring_full:

	wait_room_in_shared_ring(i);
	if (i) goto send_bcast_back;  /*  check again for available room  */
	if (is_bcast) {
/*
 *  Take note that the operation was suspended while bcasting in the
 *  same node.
 */
		out_p->bcast_was_susp = 1;

DEBUG(printg("gamma_send_2p_flowctl_samenode(): suspend bcast.\n"));

	}

DEBUG_E(printg(
"gamma_send_2p_flowctl_samenode(): head %ld,tail %ld: No room.  Try again!\n",
shared_ring.head,shared_ring.tail
));

	return -GAMMA_ERR_TRYAGAIN;

no_partner:

	return -GAMMA_ERR_GENERIC;

}


/*
 *  Returns:
 *
 *  P | IS_SAME_NODE (0 <= P <= MAX_PORT_NUMBER) if msg complete for port P
 *  (user-level caller will perform notification on sem #P and run rec handler
 *  if present; msg lenght is in user-level variable "gamma_msglen")
 *
 *  P | IS_FIRST_CHUNK | IS_SAME_NODE if msg not complete for port P but
 *  rec handler must run to manage the arrival of 1st piece of msg
 *  (user-level caller will run the handler, 1st piece lenght is in user-level
 *  variable "gamma_msglen")
 *
 *  -GAMMA_RX_INCOMPLETE if no complete msg and no handler to run, or
 *  the invocation was bogus (no messages at all)
 *
 *  -GAMMA_ERR_MUSTYIELD if there is a pending pkt for a process other than the
 *  invoker, or another process is expected to shortly be in this routine
 *
 */
static int rx_gamma_samenode
(void)
{

    int					res;
    unsigned long			bcast_sequence;
    shared_ring_entry			*desc;
    unsigned long			head, par_pid, procnum, port, cnk_len;
    unsigned long			bitmap, num_locdest, first_piece_size;
    port_entry				*pe;
    task_t				*receiver, *p;
    unsigned long			next_procn = 0;
    void				*rx_buffer;
    time_586				cur_time;


    if (gamma_must_yield_samenode.task != NULL) goto check_yield;

ok_yield:

    if (gamma_run_handler_samenode != NULL) goto rec_reentering;

DEBUG_E(printg("rx_gamma_samenode(), pid %u: entering...\n",current->pid));

/* shared ring might have been consumed by another CPU meanwhile...? */
    if (shared_ring.head == shared_ring.tail) goto bogus_inv;

    res = -GAMMA_RX_INCOMPLETE;
    bcast_sequence = 0;

    for (;;) {

        head = shared_ring.head;
        desc = &shared_ring.desc[head];

DEBUG_E(printg(
"rx_gamma_samenode(): loop: head %ld, tail %ld, bitmap %ld\n",
head,shared_ring.tail,shared_ring.desc[head].bitmap
));

        par_pid = desc->par_pid;
        procnum = desc->procnum;
        port = desc->port;
        cnk_len = desc->cnk_len;
        bitmap = desc->bitmap;
        num_locdest = desc->num_locdest;

DEBUG(printg(
"rx_gamma_samenode(), pid %u: par_pid %ld, procnum %ld, port %ld, cnk_len %ld\n",
current->pid,par_pid,procnum,port,cnk_len
));

/*  Check whether virtual GAMMA #par_pid exists and is assigned.  */
        if ((par_proc_table[par_pid] == NULL)
            || (par_proc_table[par_pid]->inst_table[procnum].pid == -1)
        )
            goto discard_pkt;

        pe = &par_proc_table[par_pid]->inst_table[procnum].port_table[port];

/*  Check whether port is in use as an input one.  */
        receiver = pe->task;
        if (receiver == NULL) goto discard_pkt;

/*  Check whether receiver is the caller.  */
        if (receiver->mm != current->mm) goto cannot_deliver_yet;

/* Check whether a seq of bcast pkts is beginning or has ended */
        if ((num_locdest > 1) && !bcast_sequence) goto found_bcast_chunk;
        if ((num_locdest == 1) && bcast_sequence) {
/*
 *  End of bcast pkt sequence.
 *  Quit the routine, to allow other receivers to consume the bcast pkt seq.
 *  NOTE: variable "gamma_must_yield_samenode", indicating which proc instance
 *  should run next in the bcast recv, has been set up when the bcast seq
 *  begun.
 */

DEBUG(printg(
"rx_gamma_samenode(): exiting (end of bcast seq), head %ld (%ld), tail %ld\n",
shared_ring.head,head,shared_ring.tail
));

            goto yield;
        }

found_bcast_chunk_back:

/*  Empty messages:  no buffer, no frame numbering.  */
        if (!cnk_len) goto end_recv_samenode;

/*  Begin of payload transfer.  */

        pe->current_rx_msglen += cnk_len;

/*
 *  Is this packet to be scattered into two pieces?
 *  If present, 1st piece is carried by 1st packet of the msg.
 */

        first_piece_size = 0;

        if (bitmap & IS_1ST_CNK) {
            first_piece_size = pe->boundary_1p;
            if ((first_piece_size > 0) && (first_piece_size <= cnk_len)) {
/*  Payload is to be scattered into two pieces.  Download 1st piece.  */

DEBUG(printg(
"rx_gamma_samenode():  Msg to be scattered, size of first piece is %ld\n",
first_piece_size
));

                get_ptr_to_user_recv_buffer(par_pid,port,procnum,rx_buffer);
                gamma_copy_to_user(rx_buffer,desc->buf,first_piece_size);
                cnk_len -= first_piece_size;

/*
 *  Prepare to quit, to run the receiver handler of the port.
 *  Since this message is going to be scattered, we can assume
 *  that there is a handler to run.
 */

/*
 *  Tell the caller which handler to run, and len of 1st chunk of data:
 *  "intra-node msg not complete but must run handler"
 */
                res = port | IS_FIRST_CHUNK | IS_SAME_NODE;
                set_msglen(par_pid,procnum,first_piece_size);

/*
 *  Current process only is allowed to reenter later,
 *  to download remaining part of payload.
 *  Note that the user-level destination of 2nd piece may have been
 *  set/reset/changed by the receiver handler invoked above.
 *  If not, 2nd piece will overwrite 1st one at the same user-level
 *  destination address, since the macro
 *  update_ptr_to_user_recv_buffer() (see below)
 *  has not been invoked.
 */
                gamma_run_handler_samenode = current;

DEBUG(printg("Quit to run rec handler for port %ld (1st piece)\n",port));

                return res;
            }
            else
                first_piece_size = 0;
        }

get_remaining_payload:

/*
 *  Shift user-space rec buffer's current ptr forward.
 *  Check for room availability.
 *  If not enough room in rec buffer, truncate the msg chunk
 *  by setting cnk_len equal to the available room.
 *  Decrement the available room by cnk_len.
 *  The current msg chunk will be stored starting
 *  from address "rx_buffer".
 */
        update_ptr_to_user_recv_buffer(par_pid,port,procnum,cnk_len,rx_buffer);

/*  Copy the payload (2nd piece) into user space  */
        recv_payload_2nd_chunk(rx_buffer,desc->buf,0,first_piece_size,cnk_len);

end_recv_samenode:

/*  Empty messages skip a lot of things.  */
        if (!desc->cnk_len) goto notif;

        if (!(bitmap & IS_LAST_CNK)) goto msg_not_complete;

/*
 *  Do one of the following, depending on the kind of buffer attached
 *  to the recv port:
 *  case GO_BACK:
 *  reset the current position ptr of the recv buffer to the begin of
 *  the recv buffer itself.  Next msg will be stored at the begin
 *  of the buffer.
 *  case GO_AHEAD:
 *  If there is still room available in the attached buffer,
 *  leave the current position ptr untouched.
 *  Otherwise, reset the current position ptr of the recv buffer to
 *  the begin of the recv buffer itself; in this case, next msg will
 *  be stored at the begin of the buffer.
 */
        switch (pe->buffer_kind) {
            case GO_BACK:
                reset_current_ptr(par_pid,port,procnum);
                break;
            case GO_AHEAD:
                check_room_possibly_reset_current_ptr(par_pid,port,procnum);
                break;
        }

notif:

/*
 *  Prepare to quit, in order to notify msg arrival and run the recv handler
 *  (if a handler is present and did not run already)
 */
        res = port;

DEBUG_E(printg(
"Setting \"gamma_msglen\" to %d\n",
par_proc_table[par_pid]->inst_table[procnum].port_table[port].current_rx_msglen
));

        set_msglen(par_pid,procnum,pe->current_rx_msglen);

/*  Next msg starts from len = 0  */
        pe->current_rx_msglen = 0;

        if (!bcast_sequence) {
/*  Not a bcast pkt:  discard as usual.  */
            desc->bitmap = IS_EMPTY;
            incr_modulo(shared_ring.head,SHARED_RING_SIZE);
            break;
        }

/*  bcast pkt: do not discard, set next receiver, decrease no. of receivers  */
        desc->procnum = next_procn;
        desc->num_locdest--;
        break;

msg_not_complete:

        if (bcast_sequence) {
/*
 *  bcast pkt:  do not discard, set next receiver, decrease no. of receivers,
 *  continue with next pkt if any, but don't increment the head of shared ring;
 *  just increment my own local copy of head, so ring remains unchanged for the
 *  next receiver in the bcast group.
 */
            desc->procnum = next_procn;
            desc->num_locdest--;
            incr_modulo(head,SHARED_RING_SIZE);
            if (head == shared_ring.tail) {
/*
 *  No more pkts available:  quit rx loop but let next receiver to run.
 *
 *  NOTE: variable "gamma_must_yield_samenode", indicating which proc instance
 *  should run next in the bcast recv, has been set up when the bcast seq
 *  begun.
 */
                break;
            }
            continue;
        }

/*  Not a bcast pkt:  discard as usual.  */

discard_pkt:

DEBUG_E(printg(
"rx_gamma_samenode(): discarding pkt, head %ld (%ld), tail %ld\n",
shared_ring.head,head,shared_ring.tail
));

        desc->bitmap = IS_EMPTY;
        incr_modulo(shared_ring.head,SHARED_RING_SIZE);

        if (shared_ring.head == shared_ring.tail)
/*  No more pkts available:  quit rx loop.  */
            break;

DEBUG(printg("rx_gamma_samenode(): continuing loop...\n"));

    }

DEBUG(printg(
"rx_gamma_samenode(), pid %u: exiting, head %ld (%ld), tail %ld, res %d\n",
current->pid,shared_ring.head,head,shared_ring.tail,res
));

/*  Quit.  Possibly yield the CPU.  */
    if ((gamma_must_yield_samenode.task != NULL) && (res < 0))
        res = -GAMMA_ERR_MUSTYIELD;

    return res;

/*  Here all the unusual branches.  */

found_bcast_chunk:

/*
 *  Found a bcast pkt.  We suppose it is the first of a sequence.
 *  Such pkt will not be removed from the ring until all receivers
 *  have copied them into user space.  Instead, remove all the other
 *  pkts received so far (by forwarding the ring tail to the current value).
 *  Also, pre-compute the next local receiver, to yield the CPU to it.
 */
    bcast_sequence = 1;

/*  Search for next receiver.  Exclude sender from the search.  */
    for (next_procn = procnum+1; next_procn < MAX_NUM_LOCAL_INST; next_procn++)
        if ((par_proc_table[par_pid]->inst_table[next_procn].task
            != desc->sender)
            && (par_proc_table[par_pid]->inst_table[next_procn].whoami >= 0)
        )
            break;
    if (next_procn == MAX_NUM_LOCAL_INST) {
/*
 *  Next receiver not found.  Should not have happened: the sender told that
 *  more than one local receiver is to get this chunk of data
 *  (desc->num_locdest > 1), so there is something bad...
 */

DEBUG_L(printg(
"rx_gamma_samenode(): Failed to find next receiver (bcast)!!  Discard...\n"
));

        goto discard_pkt;
    }

/*  Prepare to yield CPU to next receiver in bcast group. */
    gamma_must_yield_samenode.task =
        par_proc_table[par_pid]->inst_table[next_procn].task;
    gamma_time(gamma_must_yield_samenode.time);
    goto found_bcast_chunk_back;

cannot_deliver_yet:

DEBUG(printg(
"rx_gamma_samenode(): delaying rec, receiver is %u, state is %ld\n",
receiver->pid, receiver->state
));

    gamma_time(cur_time);

    if (!desc->watch_time.hi && !desc->watch_time.lo) {
        desc->watch_time = cur_time;
/*  Quit.  Yield the CPU, but allowed to keep inspecting.   */
        goto yield;
    }

    if (gamma_time_diff_ticks(cur_time,desc->watch_time) > 100*CLOCK) {
/*
 *  The pkt is sitting in the shared ring since more than 100 usec ago.
 *  Is the receiver process alive?  NOTE: we prefer wait 100 usec then
 *  perform the liveness check on receiver, for the check itself is
 *  time-consuming and the frequent case is when the receiver is alive.
 */
        p = find_task_by_pid(receiver->pid);
        if (p == NULL) goto dead_process;
        if (receiver->state & (EXIT_ZOMBIE | EXIT_DEAD)) goto dead_process;
    }

/*  Quit.  Yield the CPU and do not inspect the pkt for a while.  */
    gamma_must_yield_samenode.task = receiver;
    gamma_time(gamma_must_yield_samenode.time);
    goto yield;

dead_process:

/*  Receiver no longer exists.  Discard pkt.  */

DEBUG(printg(
"rx_gamma_samenode(): recv proc does not exist.  Discarding packet...\n"
));

    goto discard_pkt;

bogus_inv:

DEBUG_E(printg(
"rx_gamma_samenode(): bogus invocation (race on shared ring tail)\n"
));

    return -GAMMA_RX_INCOMPLETE;

rec_reentering:

    if (gamma_run_handler_samenode != current)
/*
 *  A receiver has temporarily quit to run the rec handler, but will
 *  reenter soon.
 *  Prevent other processes to enter now.
 */
        goto yield;

/*  The receiver is reentering after running a rec handler (1sh piece).  */

DEBUG_E(printg("rx_gamma_samenode(), pid %u: reentering...\n",current->pid));

    gamma_run_handler_samenode = NULL;
    res = -GAMMA_RX_INCOMPLETE;
    head = shared_ring.head;
    desc = &shared_ring.desc[head];

DEBUG_E(printg(
"rx_gamma_samenode(): loop: head %ld, tail %ld, bitmap %ld\n",
head,shared_ring.tail,shared_ring.desc[head].bitmap
));

    par_pid = desc->par_pid;
    procnum = desc->procnum;
    port = desc->port;
    cnk_len = desc->cnk_len;
    bitmap = desc->bitmap;
    num_locdest = desc->num_locdest;
    bcast_sequence = (num_locdest > 1);

/*  Check whether virtual GAMMA #par_pid exists and is assigned.  */
    if ((par_proc_table[par_pid] == NULL)
        || (par_proc_table[par_pid]->inst_table[procnum].pid == -1)
    )
        goto discard_pkt;

    pe = &par_proc_table[par_pid]->inst_table[procnum].port_table[port];
    receiver = pe->task;
    first_piece_size = pe->boundary_1p;
    cnk_len -= first_piece_size;

DEBUG(printg(
"rx_gamma_samenode(), pid %u, reentering: par_pid %ld, procnum %ld, port %ld, cnk_len %ld\n",
current->pid,par_pid,procnum,port,cnk_len
));

    goto get_remaining_payload;

check_yield:

    if (gamma_must_yield_samenode.task == current) {
        gamma_must_yield_samenode.task = NULL;
        goto ok_yield;
    }

    gamma_time(cur_time);

    if (gamma_time_diff_ticks(cur_time,gamma_must_yield_samenode.time)
        > 100000*CLOCK) {
        p = find_task_by_pid(gamma_must_yield_samenode.task->pid);
        if (p == NULL) goto dead;
        if (gamma_must_yield_samenode.task->state & (EXIT_ZOMBIE | EXIT_DEAD))
            goto dead;

DEBUG_L(printg(
"rx_gamma_samenode(): too long a waiting.  A pkt is sitting\n"
"in the Rx queue and the destination process does not read it.\n"
"SIGURG to destination process (pid %d)\n",
gamma_must_yield_samenode.task->pid));

        force_sig(SIGURG,gamma_must_yield_samenode.task);
        gamma_time(gamma_must_yield_samenode.time);
        goto yield;

dead:

        /* receiver process is dead.  go ahead, will remove dangling pkt */
        gamma_must_yield_samenode.task = NULL;
        goto ok_yield;
    }

yield:

    return -GAMMA_ERR_MUSTYIELD;

}
#endif


/************************************************************************
 *	Blocking send routines, best-effort				*
 ************************************************************************/


long sys_gamma_send
(
	unsigned long out_port,
	unsigned long my_par_pid,
	unsigned long procnum,
	void *data,
	unsigned long len
)
{

	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void				*ddata;
	unsigned long			llen;
#endif
	l2s				*h;
	unsigned long			entry;
	unsigned long			cnk_len;
	unsigned long			frame_size;
	int				i, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	int				j, nic;
#endif
#ifdef MEASURE_OUT_TH_MEM
	time_586			ima,imb;
#endif


	if (!spin_trylock(&gamma_tx_lock)) goto locked;
	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[procnum].pid < 0)
		goto no_par_pid;

	pe = &par_proc_table[my_par_pid]->inst_table[procnum]
							.port_table[out_port];
#ifdef MORE_NICS_SAMENODE_ROUTING
	nic = pe->nic_idx;
	if (nic != BCAST_ALL_NICS) /* normal path, only one nic is used */
		tx_common_init(nic);
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_same_node) goto same_node;
#endif

DEBUG_E(printg("sys_gamma_send(): Entering\n"));

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG_E(
if (nic != BCAST_ALL_NICS)
	printg("     Send to phys node %lu through NIC %d\n",phys_dest,nic);
else
	printg("     Broadcast. Send message through all NICs.\n");
);
#endif

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata = data;
	llen = len;
#endif

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
/*  copy the precomputed header into the tx buffer  */
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(h,
			cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK);

#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data,cnk_len);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif

	for (;;) {
#ifdef MORE_NICS_SAMENODE_ROUTING
		if (unlikely(nic == BCAST_ALL_NICS)) {
/* broadcast path (used only when multiple nic support is enabled) */

#ifdef SEQCHECK
			set_send_header_frameno(h,win->tx_seq);
			win->tx_seq++;
#endif
/* luigi - warning: seqcheck and debug print are BEFORE tx_ring_is_full(). 
 * this could cause problems. */			
DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));
 
			for (j=0; j < num_gamma_devices; j++) {
				tx_common_init(j);
				
				if (tx_ring_is_full()) goto bc_ring_full;

			bc_room_in_tx_ring:
				
				tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
				tx_start();
				tx_forw_tail();
			}
		}
		else {
/* point-to-point path */
#endif
		if (tx_ring_is_full()) goto ring_full;
			
room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef MORE_NICS_SAMENODE_ROUTING
		}
#endif

#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		len -= cnk_len;
		if (likely(!len)) break;

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
/*  copy previous header into new tx buffer  */
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data = (u8 *)data+cnk_len;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
		gamma_copy_from_user((u8 *)h+HEADER_SIZE,data,cnk_len);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif
		continue;

ring_full:

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

#ifdef MORE_NICS_SAMENODE_ROUTING
bc_ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto bc_room_in_tx_ring;
#endif
	}

DEBUG(printg("sys_gamma_send(): Done!\n"));

	res = 0;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST)
		res = gamma_send_2p_samenode(my_par_pid,procnum,out_port,
							NULL,0,ddata,llen,1);
#endif
	spin_unlock(&gamma_tx_lock);
	return res;

tx_failure:

DEBUG_L(printg(
"sys_gamma_send(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST)
		res = gamma_send_2p_samenode(my_par_pid,procnum,out_port,
							NULL,0,data,len,0);
	else
		res = gamma_send_2p_samenode(my_par_pid,procnum,out_port,
							NULL,0,data,len,1);
	spin_unlock(&gamma_tx_lock);
	return res;
#endif

no_par_pid:

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


long sys_gamma_send_flowctl
(
	unsigned long out_port_and_was_susp,
	unsigned long my_par_pid,
	unsigned long procnum,
	void *data,
	unsigned long len
)
{

	unsigned long			out_port;
	unsigned long			was_susp;
	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
	unsigned long			credit;
	unsigned long			curr_credit;
	l2s				*h;
	unsigned long			entry;
	unsigned long			cnk_len;
	unsigned long			frame_size;
	int				i, done, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	int				j, nic;
#endif
#ifdef MEASURE_OUT_TH_MEM
time_586 ima,imb;
#endif


	if (!spin_trylock(&gamma_tx_lock)) goto locked;
	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[procnum].pid < 0)
		goto no_par_pid;

	out_port = out_port_and_was_susp & ~0x80000000;
	was_susp = out_port_and_was_susp & 0x80000000;

	pe = &par_proc_table[my_par_pid]->inst_table[procnum]
							.port_table[out_port];
#ifdef MORE_NICS_SAMENODE_ROUTING
	nic = pe->nic_idx;
	if (nic != BCAST_ALL_NICS) /* normal path, only one nic is used */
		tx_common_init(nic);
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_same_node) goto same_node;
	if (pe->bcast_was_susp) goto same_node_bcast_susp;
#endif

DEBUG_E(printg("sys_gamma_send_flowctl(): Entering\n"));

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG_E(
if (nic != BCAST_ALL_NICS)
	printg("     Send to phys node %lu through NIC %d\n",phys_dest,nic);
else
	printg("     Broadcast. Send message through all NICs.\n");
);
#endif

	if ((phys_dest == BROADCAST) && !was_susp) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		pe->current_tx_data1 = data;
		pe->current_tx_len1 = len;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;
	if (!curr_credit) goto no_credit;

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK);
		else
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK
						| MSG_COMPLETE_MASK);
	} else if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len);
	else
		fill_send_header_cnklen(h, cnk_len | MSG_COMPLETE_MASK);

#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data,cnk_len);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit)) {
			set_send_header_creq(h);

DEBUG_F({
 int i;
 printg("send CRED_REQ frame_no %d: curr credit %ld: ",win->tx_seq,curr_credit);
 for (i=0; i<total_num_nodes; i++)
  printg("(%d) ",atomic_read(&local_credit[i]));
 printg(".\n");
})

		}

#ifdef MORE_NICS_SAMENODE_ROUTING
		if (unlikely(nic == BCAST_ALL_NICS)) {
/* broadcast path (used only when multiple nic support is enabled) */

#ifdef SEQCHECK
			set_send_header_frameno(h,win->tx_seq);
			win->tx_seq++;
#endif
/* luigi - warning: seqcheck and debug print are BEFORE tx_ring_is_full(). 
 * this could cause problems. */			
DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));
 
			for (j=0; j < num_gamma_devices; j++) {
				tx_common_init(j);
				
				if (tx_ring_is_full()) goto bc_ring_full;

			bc_room_in_tx_ring:
				
				tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
				tx_start();
				tx_forw_tail();
			}
		}
		else {
/* point-to-point path */
#endif
		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef MORE_NICS_SAMENODE_ROUTING
		}
#endif

#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		curr_credit--;
		len -= cnk_len;
		if (!len) break;
		if (!curr_credit) goto credit_exhausted;

credit_back:

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data = (u8 *)data+cnk_len;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
		gamma_copy_from_user((u8 *)h+HEADER_SIZE,data,cnk_len);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif

		continue;

credit_exhausted:

		dec_credit(phys_dest,credit);
		refill_credit(done,phys_dest,credit,curr_credit);
		if (done) goto credit_back;

DEBUG_F(printg(
"sys_gamma_send_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

		spin_unlock(&gamma_tx_lock);
		if (!len) return -GAMMA_ERR_TRYAGAIN;
		return len;

ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

#ifdef MORE_NICS_SAMENODE_ROUTING
bc_ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto bc_room_in_tx_ring;
#endif

	}

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg("sys_gamma_send_flowctl(): Done!\n"));

	res = 0;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST)
		res = gamma_send_2p_flowctl_samenode(
				my_par_pid,procnum,out_port,NULL,0,
				pe->current_tx_data1,pe->current_tx_len1,1,0);
#endif
	spin_unlock(&gamma_tx_lock);
	return res;

no_credit:

DEBUG_E(printg("sys_gamma_send_flowctl():  No credit.  Try again!\n"));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_TRYAGAIN;

tx_failure:

DEBUG_L(printg(
"sys_gamma_send_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	dec_credit(phys_dest,credit-curr_credit);

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
		res = gamma_send_2p_flowctl_samenode(
				my_par_pid,procnum,out_port,NULL,0,
				data,len,0,was_susp);
		spin_unlock(&gamma_tx_lock);
		return res;
	}

same_node_bcast_susp:

	res = gamma_send_2p_flowctl_samenode(
			my_par_pid,procnum,out_port,NULL,0,
			data,len,1,was_susp);
	spin_unlock(&gamma_tx_lock);
	return res;
#endif

no_par_pid:

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


long sys_gamma_send_2p
(
	unsigned long out_port_and_len1,
	unsigned long my_par_pid_and_procnum,
	void *data1,
	void *data2,
	unsigned long len2
)
{

	unsigned long			my_par_pid;
	unsigned long			procnum;
	unsigned long			out_port;
	unsigned long			len1;
	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void				*ddata2;
	unsigned long			llen1;
#endif
	l2s				*h;
	unsigned long			entry;
	unsigned long			len;
	unsigned long			frame_size;
	unsigned long			cnk_len;
	int				i, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	int				j, nic;
#endif


	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	procnum = my_par_pid_and_procnum >> 16;
	out_port = out_port_and_len1>>8;
	len1 = (u8)out_port_and_len1;

	if (!spin_trylock(&gamma_tx_lock)) goto locked;
	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[procnum].pid < 0)
		goto no_par_pid;
	if (len1 > MAX_1ST_PIECE_SIZE) goto first_too_long;

	pe = &par_proc_table[my_par_pid]->inst_table[procnum]
							.port_table[out_port];
#ifdef MORE_NICS_SAMENODE_ROUTING
	nic = pe->nic_idx;
	if (nic != BCAST_ALL_NICS) /* normal path, only one nic is used */
		tx_common_init(nic);
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_same_node) goto same_node;
#endif

DEBUG_E(printg("sys_gamma_send_2p(): Entering\n"));

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG_E(
if (nic != BCAST_ALL_NICS)
	printg("     Send to phys node %lu through NIC %d\n",phys_dest,nic);
else
	printg("     Broadcast. Send message through all NICs.\n");
);
#endif

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata2 = data2;
	llen1 = len1;
#endif

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data1,len1);

	len = len1 + len2;
	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(
			h, cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK);

	gamma_copy_from_user((u8 *)h+HEADER_SIZE+len1,data2,cnk_len-len1);

	for (;;) {
#ifdef MORE_NICS_SAMENODE_ROUTING
		if (unlikely(nic == BCAST_ALL_NICS)) {
/* broadcast path (used only when multiple nic support is enabled) */

#ifdef SEQCHECK
			set_send_header_frameno(h,win->tx_seq);
			win->tx_seq++;
#endif
/* luigi - warning: seqcheck and debug print are BEFORE tx_ring_is_full(). 
 * this could cause problems. */			
DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));
 
			for (j=0; j < num_gamma_devices; j++) {
				tx_common_init(j);
				
				if (tx_ring_is_full()) goto bc_ring_full;

			bc_room_in_tx_ring:
				
				tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
				tx_start();
				tx_forw_tail();
			}
		}
		else {
/* point-to-point path */
#endif
		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef MORE_NICS_SAMENODE_ROUTING
		}
#endif

#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		len -= cnk_len;
		if (!len) break;

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data2 = (u8 *)data2+cnk_len-len1;
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		gamma_copy_from_user((u8 *)h+HEADER_SIZE,data2,cnk_len);

		continue;

ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

#ifdef MORE_NICS_SAMENODE_ROUTING
bc_ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto bc_room_in_tx_ring;
#endif
	}

DEBUG(printg("sys_gamma_send_2p(): Done!\n"));

	res = 0;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST)
		res = gamma_send_2p_samenode(my_par_pid,procnum,out_port,
						data1,llen1,ddata2,len2,1);
#endif
	spin_unlock(&gamma_tx_lock);
	return res;

tx_failure:

DEBUG_L(printg(
"sys_gamma_send_2p(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST)
		res = gamma_send_2p_samenode(my_par_pid,procnum,out_port,
						data1,len1,data2,len2,0);
	else
		res = gamma_send_2p_samenode(my_par_pid,procnum,out_port,
						data1,len1,data2,len2,1);
	spin_unlock(&gamma_tx_lock);
	return res;
#endif

first_too_long:
no_par_pid:

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


long sys_gamma_send_2p_flowctl
(
	unsigned long out_port_and_len1_and_was_susp,
	unsigned long my_par_pid_and_procnum,
	void *data2,
	unsigned long len2,
	void *data1
)
{

	unsigned long			my_par_pid;
	unsigned long			procnum;
	unsigned long			out_port;
	unsigned long			len1;
	unsigned long			was_susp;
	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
	unsigned long			credit;
	unsigned long			curr_credit;
	l2s				*h;
	unsigned long			entry;
	unsigned long			len;
	unsigned long			cnk_len;
	unsigned long			frame_size;
	int				i, done, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	int				j, nic;
#endif


	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	procnum = my_par_pid_and_procnum >> 16;
	out_port = (out_port_and_len1_and_was_susp & ~0x80000000) >> 8;
	len1 = (u8)out_port_and_len1_and_was_susp;
	was_susp = out_port_and_len1_and_was_susp & 0x80000000;

	if (!spin_trylock(&gamma_tx_lock)) goto locked;
	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[procnum].pid < 0)
		goto no_par_pid;
	if (len1 > MAX_1ST_PIECE_SIZE) goto first_too_long;

	pe = &par_proc_table[my_par_pid]->inst_table[procnum]
							.port_table[out_port];

#ifdef MORE_NICS_SAMENODE_ROUTING
	nic = pe->nic_idx;
	if (nic != BCAST_ALL_NICS) /* normal path, only one nic is used */
		tx_common_init(nic);
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_same_node) goto same_node;
	if (pe->bcast_was_susp) goto same_node_bcast_susp;
#endif

DEBUG_E(printg("sys_gamma_send_2p_flowctl(): Entering\n"));

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG_E(
if (nic != BCAST_ALL_NICS)
	printg("     Send to phys node %lu through NIC %d\n",phys_dest,nic);
else
	printg("     Broadcast. Send message through all NICs.\n");
);
#endif

	if ((phys_dest == BROADCAST) && !was_susp) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		pe->current_tx_data1 = data1;
		pe->current_tx_len1 = len1;
		pe->current_tx_data2 = data2;
		pe->current_tx_len2 = len2;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;
	if (!curr_credit) goto no_credit;

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data1,len1);

	len = len1 + len2;
	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK);
		else
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK
						| MSG_COMPLETE_MASK);
	} else if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len);
	else
		fill_send_header_cnklen(h, cnk_len | MSG_COMPLETE_MASK);

	gamma_copy_from_user((u8 *)h+HEADER_SIZE+len1,data2,cnk_len-len1);

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit)) {
			set_send_header_creq(h);

DEBUG_F({
 int i;
 printg("send CRED_REQ frame_no %d: curr credit %ld: ",win->tx_seq,curr_credit);
 for (i=0; i<total_num_nodes; i++)
  printg("(%d) ",atomic_read(&local_credit[i]));
 printg(".\n");
})

		}

#ifdef MORE_NICS_SAMENODE_ROUTING
		if (unlikely(nic == BCAST_ALL_NICS)) {
/* broadcast path (used only when multiple nic support is enabled) */

#ifdef SEQCHECK
			set_send_header_frameno(h,win->tx_seq);
			win->tx_seq++;
#endif
/* luigi - warning: seqcheck and debug print are BEFORE tx_ring_is_full(). 
 * this could cause problems. */			
DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));
 
			for (j=0; j < num_gamma_devices; j++) {
				tx_common_init(j);
				
				if (tx_ring_is_full()) goto bc_ring_full;

			bc_room_in_tx_ring:
				
				tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
				tx_start();
				tx_forw_tail();
			}
		}
		else {
/* point-to-point path */
#endif
		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef MORE_NICS_SAMENODE_ROUTING
		}
#endif

#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		curr_credit--;
		len -= cnk_len;
		if (!len) break;
		if (!curr_credit) goto credit_exhausted;

credit_back:

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data2 = (u8 *)data2+cnk_len-len1;
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		gamma_copy_from_user((u8 *)h+HEADER_SIZE,data2,cnk_len);

		continue;

credit_exhausted:

		dec_credit(phys_dest,credit);
		refill_credit(done,phys_dest,credit,curr_credit);
		if (done) goto credit_back;

DEBUG_F(printg(
"sys_gamma_send_2p_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

		spin_unlock(&gamma_tx_lock);
		if (!len) return -GAMMA_ERR_TRYAGAIN;
		return len;

ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

#ifdef MORE_NICS_SAMENODE_ROUTING
bc_ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto bc_room_in_tx_ring;
#endif

	}

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg("sys_gamma_send_2p_flowctl(): Done!\n"));

	res = 0;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST)
		res = gamma_send_2p_flowctl_samenode(
				my_par_pid,procnum,out_port,
				pe->current_tx_data1,pe->current_tx_len1,
				pe->current_tx_data2,pe->current_tx_len2,1,0);
#endif
	spin_unlock(&gamma_tx_lock);
	return res;

no_credit:

DEBUG_E(printg("sys_gamma_send_2p_flowctl():  No credit.  Try again!\n"));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_TRYAGAIN;

tx_failure:

DEBUG_L(printg(
"sys_gamma_send_2p_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	dec_credit(phys_dest,credit-curr_credit);

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
		res = gamma_send_2p_flowctl_samenode(
				my_par_pid,procnum,out_port,data1,len1,
				data2,len2,0,was_susp);
		spin_unlock(&gamma_tx_lock);
		return res;
	}

same_node_bcast_susp:

	res = gamma_send_2p_flowctl_samenode(
			my_par_pid,procnum,out_port,data1,len1,
			data2,len2,1,was_susp);
	spin_unlock(&gamma_tx_lock);
	return res;
#endif

first_too_long:
no_par_pid:

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


/************************************************************************
 *	Non-blocking send routines, best-effort				*
 ************************************************************************/


long sys_gamma_isend
(
	unsigned long out_port,
	unsigned long my_par_pid,
	unsigned long procnum,
	void *data,
	unsigned long len
)
{

	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void				*ddata;
	unsigned long			llen;
#endif
	l2s				*h;
	unsigned long			entry;
	unsigned long			frame_size;
	unsigned long			cnk_len;
	void				*paddr;
	void				*next_paddr;
	int				i, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned long			res_nic_idx;
#endif


	if (!spin_trylock(&gamma_tx_lock)) goto locked;
	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[procnum].pid < 0)
		goto no_par_pid;

	pe = &par_proc_table[my_par_pid]->inst_table[procnum]
							.port_table[out_port];
#ifdef MORE_NICS_SAMENODE_ROUTING
	res_nic_idx = pe->nic_idx;
	tx_common_init(res_nic_idx);
	/* nic_idx is at most 5 bit long */
	res_nic_idx = res_nic_idx << 26;	
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_same_node) goto same_node;
#endif

DEBUG_E(printg("sys_gamma_isend(): Entering\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata = data;
	llen = len;
#endif

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data);
	if (cnk_len >= PAGE_SIZE-((unsigned long)paddr & 0x00000fff))
		goto next_page1;
	next_paddr = (u8 *)paddr+cnk_len;

next_page_back1:

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(
			h, cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK);

	for (;;) {

		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE,paddr,cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE;

		len -= cnk_len;
		if (!len) break;

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data = (u8 *)data+cnk_len;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= PAGE_SIZE-((unsigned long)paddr & 0x00000fff))
			goto next_page;
		next_paddr = (u8 *)next_paddr+cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		continue;

next_page:

		cnk_len = PAGE_SIZE-((unsigned long)paddr & 0x00000fff);
		next_paddr = user_space_to_phys(current,(u8 *)data+cnk_len);
		goto next_page_back;

ring_full:

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

	}

DEBUG(printg("sys_gamma_isend(): Done!  isend_handle %d\n",tx_ring_tail));

	res = tx_ring_tail;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST) {
		res = gamma_send_2p_samenode(my_par_pid,procnum,out_port,
						NULL,0,ddata,llen,1);
		if (!res) res = tx_ring_tail;
	}
#endif
	spin_unlock(&gamma_tx_lock);

#ifdef MORE_NICS_SAMENODE_ROUTING
	/* assuming 26 bit is enough for tx_ring_tail */
	/* luigi warning: stiamo facendo l'OR anche quando res contiene un codice di errore? 
	 * RIGUARDARE! e se lo facessimo con una macro??? */
	return res | res_nic_idx;
#else
	return res;
#endif

next_page1:

	cnk_len = PAGE_SIZE-((unsigned long)paddr & 0x00000fff);
	next_paddr = user_space_to_phys(current,(u8 *)data+cnk_len);
	goto next_page_back1;

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST)
		res = gamma_send_2p_samenode(my_par_pid,procnum,out_port,
							NULL,0,data,len,0);
	else
		res = gamma_send_2p_samenode(my_par_pid,procnum,out_port,
							NULL,0,data,len,1);
	spin_unlock(&gamma_tx_lock);
	if (!res) return -GAMMA_ISEND_IS_DONE;
	return res;
#endif

no_par_pid:

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


long sys_gamma_isend_flowctl
(
	unsigned long out_port_and_was_susp,
	unsigned long my_par_pid,
	unsigned long procnum,
	void *data,
	unsigned long len
)
{

	unsigned long			out_port;
	unsigned long			was_susp;
	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
	unsigned long			credit;
	unsigned long			curr_credit;
	l2s				*h;
	unsigned long			entry;
	unsigned long			frame_size;
	unsigned long			cnk_len;
	void				*paddr;
	void				*next_paddr;
	int				i, done, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned long			res_nic_idx;
#endif


	if (!spin_trylock(&gamma_tx_lock)) goto locked;
	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[procnum].pid < 0)
		goto no_par_pid;

	out_port = out_port_and_was_susp & ~0x80000000;
	was_susp = out_port_and_was_susp & 0x80000000;
	pe = &par_proc_table[my_par_pid]->inst_table[procnum]
							.port_table[out_port];

#ifdef MORE_NICS_SAMENODE_ROUTING
	res_nic_idx = pe->nic_idx;
	tx_common_init(res_nic_idx);
	/* nic_idx is at most 5 bit long */
	res_nic_idx = res_nic_idx << 26;
	/* luigi warning: fare i return col corretto impacchettamento... 
	 * perche' si fa store/retrieve dell'handle nella porta??? */
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_same_node) goto same_node;
	if (pe->bcast_was_susp) goto same_node_bcast_susp;
#endif

DEBUG_E(printg("sys_gamma_isend_flowctl(): Entering\n"));

	if ((phys_dest == BROADCAST) && !was_susp) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		pe->current_tx_data1 = data;
		pe->current_tx_len1 = len;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;
	if (!curr_credit) goto no_credit;

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data);
	if (cnk_len >= (PAGE_SIZE-((unsigned long)paddr & 0x00000fff)))
		goto next_page1;
	next_paddr = (u8 *)paddr+cnk_len;

next_page_back1:

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK);
		else
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK
						| MSG_COMPLETE_MASK);
	} else if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len);
	else
		fill_send_header_cnklen(h, cnk_len | MSG_COMPLETE_MASK);

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit))
			set_send_header_creq(h);

		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE,paddr,cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE;

		curr_credit--;
		len -= cnk_len;
		if (!len) break;
		if (!curr_credit) goto credit_exhausted;

credit_back:

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data = (u8 *)data+cnk_len;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= (PAGE_SIZE-((unsigned long)paddr & 0x00000fff)))
			goto next_page;
		next_paddr = (u8 *)next_paddr+cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		continue;

next_page:

		cnk_len = PAGE_SIZE-((unsigned long)paddr & 0x00000fff);
		next_paddr = user_space_to_phys(current,(u8 *)data+cnk_len);
		goto next_page_back;

credit_exhausted:

		dec_credit(phys_dest,credit);
		refill_credit(done,phys_dest,credit,curr_credit);
		if (done) goto credit_back;

DEBUG_F(printg(
"sys_gamma_isend_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

/*  Return the total lenght of the not yet sent part of msg  */
		spin_unlock(&gamma_tx_lock);
		return -(GAMMA_ERR_MAX+len);

ring_full:

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

	}

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg(
"sys_gamma_isend_flowctl(): Done!  isend_handle %d\n",tx_ring_tail
));

	res = tx_ring_tail;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST) {
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
		res = gamma_send_2p_flowctl_samenode(
				my_par_pid,procnum,out_port,NULL,0,
				pe->current_tx_data1,pe->current_tx_len1,1,0);
		if (!res)
			res = tx_ring_tail;
		else {
/*
 *  Store the inter-node isend handle: will be returned to the caller as soon
 *  as the intra-node Tx terminates.
 */
			pe->suspended_isend_handle = tx_ring_tail;
			if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
				res = -(GAMMA_ERR_MAX+len);
			else if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
				res = -(GAMMA_ERR_MAX+res);
		}
	}
#endif
	spin_unlock(&gamma_tx_lock);
	return res;

next_page1:

	cnk_len = PAGE_SIZE-((unsigned long)paddr & 0x00000fff);
	next_paddr = user_space_to_phys(current,(u8 *)data+cnk_len);
	goto next_page_back1;

no_credit:

DEBUG_E(printg("sys_gamma_isend_flowctl():  No credit.  Try again!\n"));

/*  Return the total lenght of the not yet sent part of msg  */
	spin_unlock(&gamma_tx_lock);
	return -(GAMMA_ERR_MAX+len);

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	dec_credit(phys_dest,credit-curr_credit);

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
		res = gamma_send_2p_flowctl_samenode(
			my_par_pid,procnum,out_port,NULL,0,data,len,
			0,was_susp);
		if (!res)
			res = -GAMMA_ISEND_IS_DONE;
		else if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
			res = -(GAMMA_ERR_MAX+len);
		else if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
			res = -(GAMMA_ERR_MAX+res);
		spin_unlock(&gamma_tx_lock);
		return res;
	}

same_node_bcast_susp:

/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
	res = gamma_send_2p_flowctl_samenode(
		my_par_pid,procnum,out_port,NULL,0,data,len,
		1,was_susp);
	if (!res) {
/*
 *  The intra-node Tx has finished.
 *  Recover the inter-node isend handle, return it to the caller.
 */
		res = pe->suspended_isend_handle;
		pe->suspended_isend_handle = -GAMMA_ISEND_IS_DONE;
		spin_unlock(&gamma_tx_lock);
	} else if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
		res = -(GAMMA_ERR_MAX+len);
	else if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
		res = -(GAMMA_ERR_MAX+res);
	spin_unlock(&gamma_tx_lock);
	return res;
#endif

no_par_pid:

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


long sys_gamma_isend_2p
(
	unsigned long out_port_and_len1,
	unsigned long my_par_pid_and_procnum,
	void *data1,
	void *data2,
	unsigned long len2
)
{

	unsigned long my_par_pid;
	unsigned long procnum;
	unsigned long out_port;
	unsigned long len1;
	port_entry *pe;
	unsigned long phys_dest;
	tx_tmp_ring_t *win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void *ddata2;
	unsigned long llen1;
#endif
	l2s *h;
	unsigned long entry;
	unsigned long len;
	unsigned long frame_size;
	unsigned long cnk_len;
	void *paddr;
	void *next_paddr;
	int i;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned long res_nic_idx;
#endif
	int res;


	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	procnum = my_par_pid_and_procnum >> 16;
	out_port = out_port_and_len1 >> 8;
	len1 = (u8)out_port_and_len1;

	if (!spin_trylock(&gamma_tx_lock)) goto locked;
	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[procnum].pid < 0)
		goto no_par_pid;
	if (len1 > MAX_1ST_PIECE_SIZE) goto first_too_long;

	pe = &par_proc_table[my_par_pid]->inst_table[procnum]
							.port_table[out_port];

#ifdef MORE_NICS_SAMENODE_ROUTING
	res_nic_idx = pe->nic_idx;
	tx_common_init(res_nic_idx);
	/* nic_idx is at most 5 bit long */
	res_nic_idx = res_nic_idx << 26;
	/* luigi warning: fare i return col corretto impacchettamento... */
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_same_node) goto same_node;
#endif

DEBUG_E(printg("sys_gamma_isend_2p(): Entering\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata2 = data2;
	llen1 = len1;
#endif

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	if (len1 <= MAX_1ST_PIECE_SIZE) {

	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data1,len1);

	len = len1+len2;
	frame_size = tx_optimal_frag(len1+len2);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data2);
	if (cnk_len >= len1+(PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff)))
		goto next_page1;
	next_paddr = paddr+cnk_len-len1;

next_page_back1:

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(
			h, cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK);

	for (;;) {

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE+len1,paddr,cnk_len-len1);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE+len1;

		len -= cnk_len;
		if (!len)
			break;

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data2 = (u8 *)data2+cnk_len-len1;
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= (PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff)))
			goto next_page;
		next_paddr += cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		continue;

next_page:

		cnk_len = PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff);
		next_paddr = user_space_to_phys(current,(u8 *)data2+cnk_len);
		goto next_page_back;

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

	}

DEBUG(printg("sys_gamma_isend_2p(): Done!  isend_handle %d\n",tx_ring_tail));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST) {
		spin_unlock(&gamma_tx_lock);
		return tx_ring_tail;
	}
	res = gamma_send_2p_samenode(
		my_par_pid,procnum,out_port,data1,llen1,ddata2,len2,1
	);
	if (!res)
		res = tx_ring_tail;
#else
	res = tx_ring_tail;
#endif
	spin_unlock(&gamma_tx_lock);
	return res;

next_page1:

	cnk_len = len1+PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff);
	next_paddr = user_space_to_phys(current,(u8 *)data2+cnk_len-len1);
	goto next_page_back1;

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend_2p(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

	}

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
		res = gamma_send_2p_samenode(
			my_par_pid,procnum,out_port,data1,len1,data2,len2,0
		);
		spin_unlock(&gamma_tx_lock);
		if (!res)
			return -GAMMA_ISEND_IS_DONE;
		return res;
	}
	res = gamma_send_2p_samenode(
		my_par_pid,procnum,out_port,data1,len1,data2,len2,1
	);
	spin_unlock(&gamma_tx_lock);
	if (!res)
		return -GAMMA_ISEND_IS_DONE;
	return res;
#endif

first_too_long:
no_par_pid:

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


long sys_gamma_isend_2p_flowctl
(
	unsigned long out_port_and_len1_and_was_susp,
	unsigned long my_par_pid_and_procnum,
	void *data2,
	unsigned long len2,
	void *data1
)
{

	unsigned long my_par_pid;
	unsigned long procnum;
	unsigned long out_port;
	unsigned long len1;
	unsigned long len;
	unsigned long was_susp;
        port_entry *pe;
	unsigned long phys_dest;
	tx_tmp_ring_t *win;
	unsigned long credit;
	unsigned long curr_credit;
	l2s *h;
	unsigned long entry;
	unsigned long frame_size;
	unsigned long cnk_len;
	void *paddr;
	void *next_paddr;
	int i, done;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned long res_nic_idx;
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res;
#endif


	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	procnum = my_par_pid_and_procnum >> 16;
	out_port = (out_port_and_len1_and_was_susp & ~0x80000000) >> 8;
	len1 = (u8)out_port_and_len1_and_was_susp;
	len = len1+len2;
	was_susp = out_port_and_len1_and_was_susp & 0x80000000;

	if (!spin_trylock(&gamma_tx_lock)) goto locked;
	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[procnum].pid < 0)
		goto no_par_pid;
	if (len1 > MAX_1ST_PIECE_SIZE) goto first_too_long;

	pe = &par_proc_table[my_par_pid]->inst_table[procnum]
							.port_table[out_port];

#ifdef MORE_NICS_SAMENODE_ROUTING
	res_nic_idx = pe->nic_idx;
	tx_common_init(res_nic_idx);
	/* nic_idx is at most 5 bit long */
	res_nic_idx = res_nic_idx << 26;
	/* luigi warning: fare i return col corretto impacchettamento... */
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_same_node) goto same_node;
	if (pe->bcast_was_susp) goto same_node_bcast_susp;
#endif

DEBUG_E(printg("sys_gamma_isend_2p_flowctl(): Entering\n"));

	if ((phys_dest == BROADCAST) && !was_susp) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		pe->current_tx_data1 = data1;
		pe->current_tx_len1 = len1;
		pe->current_tx_data2 = data2;
		pe->current_tx_len2 = len2;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;

	if (len1 <= MAX_1ST_PIECE_SIZE) {

	len = len1+len2;

	if (!curr_credit)
		goto no_credit;

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data1,len1);

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data2);
	if (cnk_len >= len1+(PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff)))
		goto next_page1;
	next_paddr = paddr+cnk_len-len1;

next_page_back1:

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(
				h, cnk_len | MSG_1ST_FRAME_MASK);
		else
			fill_send_header_cnklen(
				h,
				cnk_len | MSG_1ST_FRAME_MASK
				| MSG_COMPLETE_MASK);
	} else if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len);
	else
		fill_send_header_cnklen(h, cnk_len | MSG_COMPLETE_MASK);

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit))
			set_send_header_creq(h);

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE+len1,paddr,cnk_len-len1);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE+len1;

		curr_credit--;
		len -= cnk_len;
		if (!len)
			break;

		if (!curr_credit)
			goto credit_exhausted;

credit_back:

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data2 = (u8 *)data2+cnk_len-len1;
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= (PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff)))
			goto next_page;
		next_paddr += cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		continue;

next_page:

		cnk_len = PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff);
		next_paddr = user_space_to_phys(current,(u8 *)data2+cnk_len);
		goto next_page_back;

credit_exhausted:

		dec_credit(phys_dest,credit);
		refill_credit(done,phys_dest,credit,curr_credit);
		if (done)
			goto credit_back;

DEBUG_F(printg(
"sys_gamma_isend_2p_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

/*  Return the total lenght of the not yet sent part of msg  */
		spin_unlock(&gamma_tx_lock);
		return -(GAMMA_ERR_MAX+len);

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

	}

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg(
"sys_gamma_isend_2p_flowctl(): Done!  isend_handle %d\n",tx_ring_tail
));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST) {
		spin_unlock(&gamma_tx_lock);
		return tx_ring_tail;
	}
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
	res = gamma_send_2p_flowctl_samenode(
		my_par_pid,procnum,out_port,
		pe->current_tx_data1,pe->current_tx_len1,
		pe->current_tx_data2,pe->current_tx_len2,1,0);
	if (!res) {
		spin_unlock(&gamma_tx_lock);
		return tx_ring_tail;
	}
/*
 *  Store the inter-node isend handle: will be returned to the caller as soon
 *  as the intra-node Tx terminates.
 */
	pe->suspended_isend_handle = tx_ring_tail;

	spin_unlock(&gamma_tx_lock);

	if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+len);
	if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+res);
	return res;
#else
	spin_unlock(&gamma_tx_lock);
	return tx_ring_tail;
#endif

next_page1:

	cnk_len = len1+PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff);
	next_paddr = user_space_to_phys(current,(u8 *)data2+cnk_len-len1);
	goto next_page_back1;

no_credit:

DEBUG_E(printg("sys_gamma_isend_2p_flowctl():  No credit.  Try again!\n"));

/*  Return the total lenght of the not yet sent part of msg  */
	spin_unlock(&gamma_tx_lock);
	return -(GAMMA_ERR_MAX+len);

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend_2p_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	dec_credit(phys_dest,credit-curr_credit);

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

	}

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
		res = gamma_send_2p_flowctl_samenode(
			my_par_pid,procnum,out_port,data1,len1,data2,len2,0,
			was_susp
		);
		spin_unlock(&gamma_tx_lock);
		if (!res)
			return -GAMMA_ISEND_IS_DONE;
		if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
			return -(GAMMA_ERR_MAX+len);
		if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
			return -(GAMMA_ERR_MAX+res);
		return res;
	}

same_node_bcast_susp:

/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
	res = gamma_send_2p_flowctl_samenode(
		my_par_pid,procnum,out_port,data1,len1,data2,len2,1,
		was_susp);
	if (!res) {
/*
 *  The intra-node Tx has finished.
 *  Recover the inter-node isend handle, return it to the caller.
 */
		res = pe->suspended_isend_handle;
		pe->suspended_isend_handle = -GAMMA_ISEND_IS_DONE;
		spin_unlock(&gamma_tx_lock);
		return res;
	}
	spin_unlock(&gamma_tx_lock);
	if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+len);
	if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+res);
	return res;
#endif

first_too_long:
no_par_pid:

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


long sys_gamma_wsend
(unsigned long isend_handle)
{

	int i;
	tx_common_decl;


DEBUG(printg("sys_gamma_wsend(): Entering\n"));

	spin_lock(&gamma_tx_lock);

#ifdef MORE_NICS_SAMENODE_ROUTING
	tx_common_init( routing_table[ (isend_handle & 0x7c000000) >> 26 ] );
	isend_handle &= 0x3ffffff;
#else
	tx_common_init(0);
#endif

	if (isend_handle > tx_ring_tail)
		goto ring_flushed;
	if ((tx_ring_head < tx_ring_tail) && (isend_handle <= tx_ring_head))
		goto ring_flushed;

	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i);

	if (i) {

DEBUG(printg("sys_gamma_wsend(): Done\n"));

		spin_unlock(&gamma_tx_lock);
		return 0;
	}

#if 0
	tx_start();

	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i);

	if (i) {

DEBUG_L(printg("sys_gamma_wsend(): Done, but had to wake up the NIC\n"));

		spin_unlock(&gamma_tx_lock);
		return 0;
	}
#endif

DEBUG_L(printg(
"sys_gamma_wsend(): Failure (isend_handle %ld, entry %d)\n",
isend_handle,tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

ring_flushed:

DEBUG(printg("sys_gamma_wsend(): Done (tx ring flushed)\n"));

	spin_unlock(&gamma_tx_lock);
	return 0;

}


long sys_gamma_tsend
(unsigned long isend_handle)
{

	int i;
	tx_common_decl;


DEBUG(printg("sys_gamma_tsend(): Entering\n"));

	spin_lock(&gamma_tx_lock);

#ifdef MORE_NICS_SAMENODE_ROUTING
	tx_common_init( routing_table[ (isend_handle & 0x7c000000) >> 26 ] );
	isend_handle &= 0x3ffffff;
#else
	tx_common_init(0);
#endif

	if (isend_handle > tx_ring_tail)
		goto ring_flushed;
	if ((tx_ring_head < tx_ring_tail) && (isend_handle <= tx_ring_head))
		goto ring_flushed;

/* peppe da fare (test invece di flush) */
	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i);

	if (i) {

DEBUG(printg("sys_gamma_tsend(): Done\n"));

		spin_unlock(&gamma_tx_lock);
		return 1;
	}

#if 0
	tx_start();

	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i);

	if (i) {

DEBUG_L(printg("sys_gamma_tsend(): Done, but had to wake up the NIC\n"));

		spin_unlock(&gamma_tx_lock);
		return 1;
	}
#endif

DEBUG_L(printg(
"sys_gamma_tsend(): Failure (isend_handle %ld, entry %d)\n",
isend_handle,tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

ring_flushed:

DEBUG(printg("sys_gamma_tsend(): Done (tx ring flushed)\n"));

	spin_unlock(&gamma_tx_lock);
	return 1;

}


/************************************************************************
 *	Receive routines						*
 ************************************************************************/


/*
 *  Returns:
 *  P (0 <= P <= MAX_PORT_NUMBER) if msg complete for port P
 *  (user-level caller will perform notification on sem #P and run rec handler
 *  if present; msg lenght is in user-level variable "gamma_msglen")
 *
 *  P | IS_FIRST_CHUNK if msg not complete for port P but
 *  rec handler must run to manage the arrival of 1st piece of msg
 *  (user-level caller will run the handler, 1st piece lenght is in user-level
 *  variable "gamma_msglen")
 *
 *  -GAMMA_RX_INCOMPLETE if no complete msg and no handler to run, or bogus
 *  invocation (no messages at all)
 *
 *  -GAMMA_ERR_MUSTYIELD if there is a pending pkt for a process other than the
 *  invoker, or another process is expected to re-enter this routine soon.
 *
 */
int rx_gamma
(struct net_device *dev)
{

l2s *h;
unsigned long cnk_len;
unsigned long frame_no;
unsigned long phys_sender;
unsigned long par_pid;
unsigned long port;
unsigned long procnum;
task_t *receiver, *p;
int res;
#ifdef MORE_INSTANCES_ON_SAME_NODE
unsigned long bcast_sequence;
#endif
unsigned long first_piece_size;
void *rx_buffer;
#ifdef MORE_INSTANCES_ON_SAME_NODE
unsigned long bcast_phys_sender = 0;
unsigned long head = 0;
unsigned long bcast_frame_no = 0;
#endif
#ifdef SEQCHECK
u32 err_msg[4];
#endif
time_586 cur_time;
rx_common_decl_and_init;
#ifdef MEASURE_IN_TH
unsigned long cumul_len;
time_586 a,b;
#endif
#ifdef MEASURE_IN_TH_MEM
time_586 ima,imb;
#endif


#ifdef MEASURE_IN_TH
cumul_len = 0;
gamma_time(a);
#endif

if (gamma_must_yield.task != NULL) goto check_yield;

ok_yield:

#ifdef MORE_INSTANCES_ON_SAME_NODE
bcast_sequence = 0;
#endif

if (gamma_run_handler != NULL) goto rec_reentering;

DEBUG_E(printg("rx_gamma(), pid %u: entering...\n",current->pid));

rx_read_pkt_info();

if (rx_err_pktnotavail()) goto bogus_inv;

res = -GAMMA_RX_INCOMPLETE;

for (;;) {

DEBUG_E(printg(
"rx_gamma(): loop: rx_ring_head %d, rx_ring_tail %d, rx_status %4.4x\n",
rx_ring_head,rx_ring_tail,rx_status()
));

 rx_header(h);

DEBUG_E(printg(
"rx_gamma(): %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x\n",
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll, h[5].ll, h[6].ll, h[7].ll, h[8].ll
));

/*  Interpret the header  */

 cnk_len = rec_header_cnk_len(h);
 frame_no = rec_header_frame_no(h);
 phys_sender = rec_header_phys_sender(h);
 rec_header_parpid_port(h,par_pid,port,procnum);

 if (rec_header_frmtype(h) != ETH_GAMMA) goto pkt_is_not_gamma;

DEBUG(printg(
"rx_gamma(), pid %u: complete %d, par_pid %ld, procnum %ld, port %ld, cnk_len %ld, frame_no %ld, phys_sender %ld, creq %d\n",
current->pid,
rec_header_msg_complete(h), par_pid, procnum, port, cnk_len, frame_no,
phys_sender, rec_header_creq(h)
));

#ifdef SEQCHECK
/****************************************************************************/
/*  Are we refusing regular GAMMA pkts from phys_sender?  */
 if (rx_discard[phys_sender]) goto end_rx;
/****************************************************************************/
#endif

accept_pkt:

#ifdef MEASURE_IN_TH
 cumul_len += cnk_len;
#endif

#ifdef SEQCHECK
/****************************************************************************/
/*  Check for correct frame sequence numbering.  */
 if (rec_is_bcast(h)) {
/*  Bcast pkts have distinct numbering.  */
  if (frame_no != rx_seq_bcast[phys_sender]) goto err_frame_no;
  rx_seq_bcast[phys_sender]++;
 } else {
  if (frame_no != rx_seq[phys_sender]) goto err_frame_no;
  rx_seq[phys_sender]++;
 }
/****************************************************************************/
#endif

/*  GAMMA pkts not allowed if parallel job not started  */
 if (!gamma_susp_irq_count) goto not_assigned;

/*  Check whether virtual GAMMA #par_pid exists and is assigned.  */
 if (
  (par_proc_table[par_pid] == NULL)
  || (par_proc_table[par_pid]->inst_table[procnum].pid == -1)
 )
  goto not_assigned;

/*  Check whether port is in use as an input one.  */
 receiver = par_proc_table[par_pid]->inst_table[procnum].port_table[port].task;
 if (receiver == NULL) goto not_assigned;

/*  Check whether receiver is the caller.  */
 if (receiver->mm != current->mm) goto cannot_deliver_yet;

/*  Flow Control:  manage the request for credit renewal, if present.  */
 if (rec_header_creq(h)) goto credit_request;

credit_request_back:

#ifdef MORE_INSTANCES_ON_SAME_NODE
/* Check whether a seq of bcast pkts is beginning or has ended */
 if (rec_is_bcast(h) && (!bcast_sequence)) goto found_bcast_pkt;
 if (bcast_sequence &&
  ((!rec_is_bcast(h)) || (phys_sender != bcast_phys_sender))) {
/*
 *  A seq. of bcast pkts has ended.
 *  Quit the routine, to allow other receivers to get the bcast data
 *  before considering the current pkt.
 */

DEBUG(printg("rx_gamma(): end of bcast pkt seq\n"));

/*
 *  NOTE: variable "gamma_must_yield", indicating which proc instance should
 *  run next in the bcast recv, has been set up when the bcast seq begun.
 */
  goto exit;
 }

found_bcast_pkt_back:
#endif

/*  Frame is OK and we get it now.  */

/*  Empty messages:  nothing to download.  */
 if (!cnk_len) goto notif;

/*  Download the payload.  */

 par_proc_table[par_pid]->inst_table[procnum].port_table[port]
 .current_rx_msglen += cnk_len;

/*
 *  Is this packet to be scattered into two pieces?
 *  If present, 1st piece is carried by 1st packet of the msg.
 */

 first_piece_size = 0;

 if (rec_header_1st_frame(h)) {
  first_piece_size =
   par_proc_table[par_pid]->inst_table[procnum].port_table[port].boundary_1p;
  if ((first_piece_size > 0) && (first_piece_size <= cnk_len)) {
/*  Payload is to be scattered into two pieces.  */

DEBUG(printg(
"rx_gamma():  Msg to be scattered, size of first piece is %ld\n",
first_piece_size
));

   get_ptr_to_user_recv_buffer(par_pid,port,procnum,rx_buffer);

DEBUG_E(printg("rx_gamma():  Copying 1st chunk to addr %p\n",rx_buffer));

   recv_payload_1st_chunk(rx_buffer,h,HEADER_SIZE,first_piece_size);
   cnk_len -= first_piece_size;

/*
 *  Prepare to quit, to run the receiver handler of the port.
 *  Since this message is going to be scattered, we can assume
 *  that there is a handler to run.
 */

/*
 *  Tell the caller which handler to run, and len of 1st chunk of data:
 *  "inter-node msg not complete but must run handler"
 */
   res = port | IS_FIRST_CHUNK;

   set_msglen(par_pid,procnum,first_piece_size);

/*
 *  Current process only is allowed to reenter later,
 *  to download remaining part of payload.
 *  Note that the user-level destination of 2nd piece may have been
 *  set/reset/changed by the receiver handler invoked above.
 *  If not, 2nd piece will overwrite 1st one at the same user-level
 *  destination address, since the macro
 *  update_ptr_to_user_recv_buffer() (see below)
 *  has not been invoked.
 */
   gamma_run_handler = current;

DEBUG(printg("Quit to run rec handler for port %ld (1st piece)\n",port));

   return res;

  } else
   first_piece_size = 0;
 }

get_remaining_payload:

/*
 *  Shift user-space rec buffer's current ptr forward.
 *  Check for room availability.
 *  If not enough room in rec buffer, truncate the msg chunk
 *  by setting cnk_len equal to the available room.
 *  Decrement the available room by cnk_len.
 *  The current msg chunk will be stored starting
 *  from address "rx_buffer".
 */
 update_ptr_to_user_recv_buffer(par_pid,port,procnum,cnk_len,rx_buffer);

#ifdef MEASURE_IN_TH_MEM
gamma_time(ima);
#endif

/*  Copy the payload (2nd piece) into user space  */
#ifdef USE_DAMNICK

DEBUG_E(printg(
"rx_gamma():  Start copying payload to addr %p through DAMNICK\n",
rx_buffer)
);

 if (!rec_header_msg_complete(h))
  damnick_icopy_k2u(
   rx_buffer,
   (void *)((u8 *)h+HEADER_SIZE+first_piece_size),
   cnk_len
  );
 else {
  unsigned long pre_notif_val;
  pre_notify_arrival(par_pid,port,procnum,pre_notif_val);
  damnick_icopy_k2u_nu(
   rx_buffer,
   (void *)((u8 *)h+HEADER_SIZE+first_piece_size),
   cnk_len,
   (unsigned long *)(par_proc_table[par_pid]->inst_table[procnum].sem1_table)
   + par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore,
   pre_notif_val
  );
 }
#else

DEBUG_E(
printg("rx_gamma():  Copying payload to addr %p, cnk_len %lu\n",
rx_buffer,cnk_len)
);

 recv_payload_2nd_chunk(rx_buffer,h,HEADER_SIZE,first_piece_size,cnk_len);
#endif

#ifdef MEASURE_IN_TH_MEM
gamma_time(imb);
in_counter_mem++;
in_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
in_total_amount_mem_bytes += cnk_len;
#endif

 if (!rec_header_msg_complete(h)) goto end_rx;

DEBUG(printg("Message complete\n"));

/*
 *  Do one of the following, depending on the kind of buffer attached
 *  to the recv port:
 *  case GO_BACK:
 *  reset the current position ptr of the recv buffer to the begin of
 *  the recv buffer itself.  Next msg will be stored at the begin
 *  of the buffer.
 *  case GO_AHEAD:
 *  If there is still room available in the attached buffer,
 *  leave the current position ptr untouched.
 *  Otherwise, reset the current position ptr of the recv buffer to
 *  the begin of the recv buffer itself; in this case, next msg will
 *  be stored at the begin of the buffer.
 */
 switch (
  par_proc_table[par_pid]->inst_table[procnum].port_table[port].buffer_kind
 ) {
  case GO_BACK:
   reset_current_ptr(par_pid,port,procnum);
   break;
  case GO_AHEAD:
   check_room_possibly_reset_current_ptr(par_pid,port,procnum);
   break;
 }

#ifdef USE_DAMNICK
/*
 *  Skip CPU-operated notification of arrival: it will be carried out by the
 *  DAMNICK device, because of the invocation of damnick_icopy_k2u_nu() .
 */
 goto set_msglen;
#endif

notif:

#ifdef USE_DAMNICK

DEBUG(printg("Notifying arrival at port %ld\n",port));

 {
 unsigned long pre_notif_val;
 pre_notify_arrival(par_pid,port,procnum,pre_notif_val);
 damnick_writel_u(
  (unsigned long *)
  (par_proc_table[par_pid]->inst_table[procnum].sem1_table)+
  par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore,
  pre_notif_val
 );
 }

set_msglen:

#endif

/*
 *  Prepare to quit, to notify msg arrival and run the receiver handler
 *  (if a handler is present and it did not run already)
 */
 res = port;

DEBUG_E(printg(
"Setting \"gamma_msglen\" to %d\n",
par_proc_table[par_pid]->inst_table[procnum].port_table[port].current_rx_msglen
));

 set_msglen(
  par_pid,procnum,
  par_proc_table[par_pid]->inst_table[procnum].port_table[port]
   .current_rx_msglen
 );

/*  Next msg starts from len = 0  */
 par_proc_table[par_pid]->inst_table[procnum].port_table[port]
 .current_rx_msglen = 0;

/*
 *  Tell the NIC this packet has been consumed, then quit rx_gamma()
 *  because a new message is now complete.
 *  Bcast pkts cannot be removed from Rx ring until all other receivers have
 *  read them.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
 if (bcast_sequence) goto quit_rx_loop_bcastseq;
#endif
 discard_pkt();
 break;

#ifdef MORE_INSTANCES_ON_SAME_NODE
quit_rx_loop_bcastseq:

 rec_header_set_procnum(h,procnum+1);
 forward_rx_ring_head();
/*
 *  NOTE: variable "gamma_must_yield", indicating which proc instance should
 *  run next in the bcast recv, has been set up when the bcast seq begun.
 */
 break;
#endif

end_rx:

/*
 *  Tell the NIC this packet has been consumed.
 *  Bcast pkts cannot be removed from Rx ring until all other receivers have
 *  read them.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
 if (!bcast_sequence)
  discard_pkt();
 else {
  rec_header_set_procnum(h,procnum+1);
  forward_rx_ring_head();
 }
#else
 discard_pkt();
#endif

end_rx_with_updated_ptrs:

 rx_read_pkt_info();

/*  If no further pending frames, quit rx_gamma()  */
 if (rx_err_pktnotavail()) break;

/*  There is at least another pending frame.  */

DEBUG_E(printg("rx_gamma(): continuing loop...\n"));

}

exit:

#ifdef MEASURE_IN_TH
gamma_time(b);
#ifdef MEASURE_IN_TH_PRINTALL
printg("in: %ld usec, %ld bytes, %ld MByte/s\n",
	((unsigned long)gamma_time_diff_ticks(b,a))/CLOCK,cumul_len,
	(cumul_len*CLOCK)/(unsigned long)gamma_time_diff_ticks(b,a)
);
#endif
in_counter++;
in_total_amount += (cumul_len*CLOCK)/(unsigned long)gamma_time_diff_ticks(b,a);
#endif

#ifdef MORE_INSTANCES_ON_SAME_NODE
if (bcast_sequence) goto quit_rx_gamma_bcastseq;
#endif

/*  Quit.  Possibly yield the CPU.  */
if (gamma_must_yield.task != NULL && res < 0)
 res = -GAMMA_ERR_MUSTYIELD;

#ifdef MORE_INSTANCES_ON_SAME_NODE
DEBUG_E(printg(
"rx_gamma(): exiting: rx_ring_head %d (%ld), rx_ring_tail %d, res %d\n",
 rx_ring_head,head,rx_ring_tail,res
));
#else
DEBUG_E(printg(
"rx_gamma(): exiting: rx_ring_head %d, rx_ring_tail %d, res %d\n",
 rx_ring_head,rx_ring_tail,res
));
#endif

return res;

#ifdef MORE_INSTANCES_ON_SAME_NODE
quit_rx_gamma_bcastseq:

/*
 *  Next local receiver must re-read bcast pkts.
 *  Set ring head back, and roll-back frame numbering sequence.
 */
write_rx_ring_head(head);
rx_seq_bcast[bcast_phys_sender] = bcast_frame_no;
/*
 *  Quit.  Yield the CPU.
 *  "gamma_must_yield" already set to next bcast receiver.
 */
if (res < 0) res = -GAMMA_ERR_MUSTYIELD;

DEBUG_E(printg(
"rx_gamma(): exiting (bcast): rx_ring_head %d (%ld), rx_ring_tail %d, res %d\n",
 rx_ring_head,head,rx_ring_tail,res
));

return res;
#endif


/*  Here all the unusual branches.  */

#ifdef SEQCHECK
err_frame_no:

/*
 *  There are two possibilities, namely:
 *  a)  A previous frame was lost.
 *      In this case, a retransmission must be asked to sender.
 *  b)  This frame is a useless duplicate of a previous one,
 *      maybe caused by a late collision on a shared LAN.
 *      In this case the previous one should have been discarded
 *      and replaced by this one, but this is no longer possible
 *      (a receiver handler could have been issued).  Since no
 *      error was detected on the previous frame, we hope that
 *      everything is OK and this duplicate may be discarded.
 *      However late collisions should never occur if the LAN is
 *      properly cabled.
 */

if (!rec_is_bcast(h)) {
 if (frame_no == (u8)(rx_seq[phys_sender]-1)) {
/*  Case b).  Discard...  */

DEBUG_L(printg(
"rx_gamma():  par_pid %ld, port %ld:  discarding duplicated frame #%ld, rx_status %8.8x\n",
par_pid,port,frame_no,rx_status()
));

  goto end_rx;
 }
} else
 if (frame_no == (u8)(rx_seq_bcast[phys_sender]-1)) {
/*  Case b).  Discard...  */

DEBUG_L(printg(
"rx_gamma():  par_pid %ld, port %ld:  discarding duplicated bcast frame #%ld, rx_status %8.8x\n",
par_pid,port,frame_no,rx_status()
));

  goto end_rx;
 }

/*  Case a).  */

DEBUG_L(
if (!rec_is_bcast(h)) {
 printg("rx_gamma(): frame #%ld, expected #%d, phys sender %ld\n",
 frame_no,rx_seq[phys_sender],phys_sender);
} else {
 printg("rx_gamma(): bcast frame #%ld, expected #%d, phys sender %ld\n",
 frame_no,rx_seq_bcast[phys_sender],phys_sender);
}
);

if (!rec_is_bcast(h)) {
 if (((frame_no-rx_seq[phys_sender]+256)%256) > MAX_CREDIT_SIZE + LOW_WATERMARK) {

DEBUG_L(printg("Out of order frame?  Discard...\n"));

  goto end_rx;
 }
} else {
 if (((frame_no-rx_seq_bcast[phys_sender]+256)%256) > MAX_CREDIT_SIZE + LOW_WATERMARK) {

DEBUG_L(printg("Out of order bcast frame?  Discard...\n"));

  goto end_rx;
 }
}

/*
 *  The interval between consecutive checks for correct pkt numbering
 *  is restored to the base value (it was growing as long as no pkt losses were
 *  observed).
 */
max_polls = base_max_polls;

/*  A missing pkt could indicate a congestion.  Ask everybody to slow down.  */
gamma_kernel_send(BROADCAST,DONT_MIND,CONGESTION,NULL,0,1);
/*
 *  Send a NACK to the origin of this pkt, to enforce retransmission of pkts
 *  starting from actual count (go-back-N).
 */
if (!rec_is_bcast(h)) {
 err_msg[0] = rx_seq[phys_sender];
 err_msg[1] = par_pid;
 err_msg[2] = port;
 err_msg[3] = procnum;
 gamma_kernel_send(
  phys_sender,DONT_MIND,NACK,err_msg,4*sizeof(u32),1
 );
} else {
 err_msg[0] = rx_seq_bcast[phys_sender];
 err_msg[1] = par_pid;
 err_msg[2] = port;
 err_msg[3] = procnum;
 gamma_kernel_send(
  phys_sender,DONT_MIND,NACKB,err_msg,4*sizeof(u32),1
 );
}
/*  We refuse other pkts from phys_sender until we get the missing ones.  */
rx_discard[phys_sender] = 1;

goto end_rx;
#endif

not_assigned:

/*
 *  Virtual GAMMA not activated or application not yet registered or port not
 *  in use or no job running.
 *  Before discarding the pkt, check for a request for a flow ctl
 *  credit renewal...
 */
if (rec_header_creq(h)) {

DEBUG_F(printg("credit request from node %ld, frame_no %ld\n",
phys_sender,frame_no))

 gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
 rec_header_clean_creq(h);
}

DEBUG(
printg("Par_pid or port not assigned, or no job running. Discarding frame (rx_status %8.8x)...\n",
rx_status())
);

goto end_rx;

pkt_is_not_gamma:

if (rec_header_frmtype(h) == ETH_GAMMA_KERNEL) {
/*
 *  GAMMA kernel msg.
 *  In this case ``procnum'' is the kind of kernel msg.
 */
#ifdef SEQCHECK
/****************************************************************************/
 if (procnum == FILL_CREDIT || procnum == CRED_REQ) {
/*  Are we refusing regular/flowctl GAMMA pkts from phys_sender?  */
  if (rx_discard[phys_sender])
   goto end_rx;
/*  Check for correct frame sequence numbering.  */
  if (rec_is_bcast(h)) {
/*  Bcast pkts have distinct numbering.  */
   if (frame_no != rx_seq_bcast[phys_sender])
    goto err_frame_no;
   rx_seq_bcast[phys_sender]++;
  } else {
   if (frame_no != rx_seq[phys_sender])
    goto err_frame_no;
   rx_seq[phys_sender]++;
  }
 }
/****************************************************************************/
#endif

DEBUG(printg(
"rx_gamma(): kernel! complete %d, parpid %ld, type %ld, port %ld, cnk_len %ld, phys_sender %ld\n",
rec_header_msg_complete(h), par_pid, procnum, port, cnk_len, phys_sender
));

DEBUG_F(
if (procnum == FILL_CREDIT)
 printg("FILL_CREDIT from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
if (procnum == CRED_REQ)
 printg("CRED_REQ from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
)

 gamma_kernel_rec((u8 *)h+HEADER_SIZE,par_pid,cnk_len,procnum,phys_sender);

 if (procnum == CRED_REQ || procnum == FILL_CREDIT || procnum == RETX_START
     || procnum == RETX_END)
  goto end_rx;
/*
 *  For the other ETH_GAMMA_KERNEL pkts, we quit rx_gamma() to avoid a possible
 *  monopolization by this thread.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
 if (!bcast_sequence)
  discard_pkt();
 else {
  rec_header_set_procnum(h,procnum+1);
  forward_rx_ring_head();
 }
#else
 discard_pkt();
#endif
 goto exit;
}
#ifdef SEQCHECK
if (rec_header_frmtype(h) == ETH_GAMMA_RETX) {
/*
 *  Regular GAMMA data pkt, simply retransmitted.
 *  Can only arrive after a ETH_GAMMA_KERNEL pkt marked as "START_RETX"
 *  (start of retransmission sequence) has arrived from node "phys_sender",
 *  so we should find rx_discard[phys_sender] == 2 (accept retrasmission
 *  sequence from node "phys_sender").
 */
/*  Are we refusing regular GAMMA pkts from phys_sender?  */
 if (rx_discard[phys_sender] == 1) {

DEBUG_F(printg(
"rx_gamma():  refusing ETH_GAMMA_RETX pkt, frame_no %ld, sender %ld\n",
frame_no,phys_sender
));

  goto end_rx;
 }
 if (rx_discard[phys_sender] == 2) {
/*  Treat this pkt as it were a regular one  */

/*  GAMMA pkts not allowed if parallel job not started  */
  if (!gamma_susp_irq_count) goto end_rx;

DEBUG_F(
printg(
"rx_gamma(), pid %u: RETX! frame %d, %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x\n",
current->pid, read_send_header_frameno(h),
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll, h[5].ll);
printg(
"complete %d, par_pid %ld, procnum %ld, port %ld, cnk_len %ld, frame_no %ld, phys_sender %ld, creq %d\n",
rec_header_msg_complete(h), par_pid, procnum, port, cnk_len, frame_no,
phys_sender, rec_header_creq(h))
);

  goto accept_pkt;
 }
/*
 *  This retransmitted pkt is unexpected.  Maybe because
 *  it is a retransmitted bcast (in this case we must discard it);
 *  or, a RETX_START pkt has been lost
 *  (and we should rather ask the sender to restart its retransmission in
 *  this case; however, the protocol will eventually recover even if we
 *  discard this pkt).
 */

DEBUG_F(printg(
"rx_gamma(): unexpected ETH_GAMMA_RETX pkt, frame_no %ld, phys_sender %ld\n",
frame_no,phys_sender
));

  goto end_rx;
}
if (rec_header_frmtype(h) == ETH_GAMMA_KERNEL_RETX) {
/*
 *  GAMMA kernel msg, retransmitted.
 *  In this case ``procnum'' is the kind of kernel msg.
 *  Possible pkt kinds are:  FILL_CREDIT,CRED_REQ.
 */
/*  Are we refusing regular GAMMA pkts from phys_sender?  */
 if (rx_discard[phys_sender] == 1) {

DEBUG_F(printg(
"rx_gamma():  refusing ETH_GAMMA_KERNEL_RETX pkt, frame_no %ld, sender %ld\n",
frame_no,phys_sender
));

  goto end_rx;
 }
 if (rx_discard[phys_sender] == 2) {
/*  Treat this pkt as it were a regular one  */
/*  Check for correct frame sequence numbering.  */
  if (rec_is_bcast(h)) {
/*  Bcast pkts have distinct numbering.  */
   if (frame_no != rx_seq_bcast[phys_sender])
    goto err_frame_no;
   rx_seq_bcast[phys_sender]++;
  } else {
   if (frame_no != rx_seq[phys_sender])
    goto err_frame_no;
   rx_seq[phys_sender]++;
  }

DEBUG_F(printg(
"rx_gamma(): retx kernel! complete %d, parpid %ld, type %ld, port %ld, cnk_len %ld, phys_sender %ld\n",
rec_header_msg_complete(h), par_pid, procnum, port, cnk_len, phys_sender
));

DEBUG_F(
if (procnum == FILL_CREDIT)
 printg("FILL_CREDIT (retx) from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
if (procnum == CRED_REQ)
 printg("CRED_REQ (retx) from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
)

  gamma_kernel_rec((u8 *)h+HEADER_SIZE,par_pid,cnk_len,procnum,phys_sender);

  if (procnum == FILL_CREDIT || procnum == RETX_START || procnum == RETX_END)
   goto end_rx;
/*
 *  For the other ETH_GAMMA_KERNEL pkts, we quit rx_gamma() to avoid a possible
 *  monopolization of this thread.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
  if (!bcast_sequence)
   discard_pkt();
  else {
   rec_header_set_procnum(h,procnum+1);
   forward_rx_ring_head();
  }
#else
  discard_pkt();
#endif
  goto exit;
 }
/*
 *  This retransmitted pkt is unexpected.  Maybe because
 *  it is a retransmitted bcast (in this case we must discard it);
 *  or, a RETX_START pkt has been lost
 *  (and we should rather ask the sender to restart its retransmission in
 *  this case; however, the protocol will eventually recover even if we
 *  discard this pkt).
 */

DEBUG_F(printg(
"rx_gamma(): unexpected ETH_GAMMA_KERNEL_RETX pkt, frame_no %ld, phys_sender %ld\n",
frame_no,phys_sender
));

 goto end_rx;
}
#endif
/*  Non-GAMMA frame.  "h" contains its first HEADER_SIZE bytes.  */

DEBUG_E(printg("rx_gamma(): non-GAMMA frame.\n"));

/*  Route the frame through the standard path if GAMMA not running.  */
if (!gamma_initialized)
 invoke_standard_rx(dev);
else
 discard_pkt();
goto end_rx_with_updated_ptrs;

#ifdef MORE_INSTANCES_ON_SAME_NODE
found_bcast_pkt:

/*
 *  Found a bcast pkt.  We suppose it is the first of a sequence.
 *  Such pkts will not be removed from the Rx ring until all receivers
 *  have copied them into user space.  Instead, remove all the other
 *  pkts received so far (by forwarding the Rx ring head to the current pos).
 *  Also, pre-compute the next local receiver, to yield the CPU to it.
 */
/*  Is there any receiver next?  */
if (
 (procnum < MAX_NUM_LOCAL_INST-1) &&
 (par_proc_table[par_pid]->inst_table[procnum+1].whoami >= 0)
) {
/*  Yes, there is!  */

DEBUG(printg("rx_gamma():  start of bcast seq\n"));

 bcast_sequence = 1;
 bcast_phys_sender = phys_sender;
/*
 *  Take note of the current Rx ring head and frame numbering:
 *  next local receiver will have to re-read pkts from there on.
 */
 read_rx_ring_head(head);
 bcast_frame_no = frame_no;

/*  Prepare to yield CPU to next receiver in the bcast group.  */
 gamma_must_yield.task = par_proc_table[par_pid]->inst_table[procnum+1].task;
 gamma_time(gamma_must_yield.time);
}
/*  If no next local receiver, treat bcast as it were unicast.  */
goto found_bcast_pkt_back;
#endif

cannot_deliver_yet:

{
 task_t *p;
/*  Is the receiver process alive?  */
 p = find_task_by_pid(receiver->pid);
 if (p == NULL) goto dead_process;
 if (receiver->state & (EXIT_ZOMBIE | EXIT_DEAD)) goto dead_process;

DEBUG(printg(
"rx_gamma(), pid %u (%p):  delaying recv of pkt #%lu, receiver is %u (%p), state is %ld\n",
current->pid,current->mm,frame_no,receiver->pid,receiver->mm,receiver->state
));

#ifdef SEQCHECK
/****************************************************************************/
/*  Rollback the frame sequence numbering.  */
 if (rec_is_bcast(h))
/*  Bcast pkts have distinct numbering.  */
  rx_seq_bcast[phys_sender]--;
 else
  rx_seq[phys_sender]--;
/****************************************************************************/
#endif
/*  Quit, leave pkt in queue. Yield the CPU.  */
 gamma_must_yield.task = receiver;
 gamma_time(gamma_must_yield.time);
 goto yield;

}

dead_process:

/*  receiver is dead.  discard packet.  */

DEBUG(printg("rx_gamma(): recv proc does not exist.  Discarding packet...\n"));

goto end_rx;

credit_request:

DEBUG_F(printg(
"credit request from node %ld, frame_no %ld (is_bcast %d)\n",
phys_sender,frame_no,rec_is_bcast(h)
))

/*  This packet carries a credit request. Refill credit to sender.  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Exception: if this is a bcast, refill credit
 *  only if this receiver is the last one of the bcast.
 */
if (!(rec_is_bcast(h))) {
#endif
/*  Refill credit and clean the credit request.  */
 gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
 rec_header_clean_creq(h);
 goto credit_request_back;
#ifdef MORE_INSTANCES_ON_SAME_NODE
}
/*
 *  Bcast packet.
 *  Refill (and clean the request) iff no other receivers (excluding this one).
 */
if (
 (procnum >= MAX_NUM_LOCAL_INST-1) ||
 (par_proc_table[par_pid]->inst_table[procnum+1].whoami < 0)
) {
 gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
 rec_header_clean_creq(h);
}
goto credit_request_back;
#endif

bogus_inv:

DEBUG_E(printg(
"rx_gamma(): bogus invocation (rx_status %8.8x), pid %u\n",
rx_status(),current->pid
));

return -GAMMA_RX_INCOMPLETE;

rec_reentering:

if (gamma_run_handler != current)
/*
 *  A receiver has temporarily quit to run the rec handler, but will
 *  reenter soon.
 *  Prevent other processes to enter now.
 */
 goto yield;

/*  The receiver is reentering after running a rec handler (1sh piece).  */

DEBUG_E(printg("rx_gamma(), pid %u: reentering...\n",current->pid));

gamma_run_handler = NULL;
rx_read_pkt_info();
res = -GAMMA_RX_INCOMPLETE;

DEBUG_E(printg(
"rx_gamma(): rx_ring_head %d, rx_ring_tail %d, rx_status %4.4x\n",
rx_ring_head,rx_ring_tail,rx_status()
));

rx_header(h);

cnk_len = rec_header_cnk_len(h);
frame_no = rec_header_frame_no(h);
phys_sender = rec_header_phys_sender(h);
rec_header_parpid_port(h,par_pid,port,procnum);

DEBUG(printg(
"rx_gamma(), pid %u:  reentering: complete %d, par_pid %ld, procnum %ld, port %ld, cnk_len %ld, frame_no %ld, phys_sender %ld, creq %d\n",
current->pid,
rec_header_msg_complete(h), par_pid, procnum, port, cnk_len, frame_no,
phys_sender, rec_header_creq(h)
));

/*  Check whether virtual GAMMA #par_pid exists and is assigned.  */
 if (
  (par_proc_table[par_pid] == NULL)
  || (par_proc_table[par_pid]->inst_table[procnum].pid == -1)
 )
  goto not_assigned;

receiver = par_proc_table[par_pid]->inst_table[procnum].port_table[port].task;

#ifdef MORE_INSTANCES_ON_SAME_NODE
bcast_sequence = 0;
if (rec_is_bcast(h)) {
 if (
  (procnum < MAX_NUM_LOCAL_INST-1) &&
  (par_proc_table[par_pid]->inst_table[procnum+1].whoami >= 0)
 ) {
  bcast_sequence = 1;
  read_rx_ring_head(head);
  bcast_frame_no = frame_no;
  gamma_must_yield.task = par_proc_table[par_pid]->inst_table[procnum+1].task;
  gamma_time(gamma_must_yield.time);
 }
}
#endif

first_piece_size =
 par_proc_table[par_pid]->inst_table[procnum].port_table[port].boundary_1p;

cnk_len -= first_piece_size;

goto get_remaining_payload;

check_yield:

if (gamma_must_yield.task == current) {
 gamma_must_yield.task = NULL;
 goto ok_yield;
}

gamma_time(cur_time);

if (gamma_time_diff_ticks(cur_time,gamma_must_yield.time) > 100000*CLOCK) {
 p = find_task_by_pid(gamma_must_yield.task->pid);
 if (p == NULL) goto dead;
 if (gamma_must_yield.task->state & (EXIT_ZOMBIE | EXIT_DEAD)) goto dead;

DEBUG_L(printg(
"rx_gamma(): too long a waiting.  A pkt is sitting\n"
"in the Rx queue and the destination process does not read it.\n"
"SIGURG to destination process (pid %d)\n", gamma_must_yield.task->pid));

 force_sig(SIGURG,gamma_must_yield.task);
 gamma_time(gamma_must_yield.time);
 goto yield;

dead:

 /* receiver dead.  go ahead, will remove dangling pkt */
 gamma_must_yield.task = NULL;
 goto ok_yield;
}

yield:

return -GAMMA_ERR_MUSTYIELD;

}


long sys_gamma_poll
(void)
{

	int res  = -GAMMA_RX_INCOMPLETE;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res2 = -GAMMA_RX_INCOMPLETE;
#endif
#ifdef MORE_NICS_SAMENODE_ROUTING
	int i;
	poll_common_decl; 
#else
	poll_common_decl_and_init;
#endif
	time_586 a,b;


	gamma_time(a);
	
	for (;;) {
/*  Check out the NIC rx ring (inter-node msgs)  */

		if (!spin_trylock(&gamma_rx_lock)) goto skip_inter;

#ifdef MORE_NICS_SAMENODE_ROUTING
/* scan all NICS sequentially and return as soon as a packet is found  */
		for (i=0; i < num_gamma_devices; i++) {
			rx_read_pkt_info_light_arg(i);
			if (!(rx_err_pktnotavail())) {
				res = rx_gamma(gamma_devices[i]);

#else
		{
			rx_read_pkt_info_light();
			if (!(rx_err_pktnotavail())) {
				res = rx_gamma(gamma_devices[0]);
#endif
				clear_rx_irqs();
#ifdef SEQCHECK
				how_many_polls = 0;
#endif
				if (likely(res >= 0)) {
/*  New msg complete, or first chunk arrived and handler must run.  */
					spin_unlock(&gamma_rx_lock);
					return res;
				}
/*  res can be:  -GAMMA_RX_INCOMPLETE, -GAMMA_ERR_MUSTYIELD  */
			}
		}

		spin_unlock(&gamma_rx_lock);

skip_inter:

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*  Check for intra-node messages  */

		if (!spin_trylock(&gamma_rxsamenode_lock)) goto skip_intra;

		if (shared_ring.head != shared_ring.tail) {
			res2 = rx_gamma_samenode();
#ifdef SEQCHECK
			how_many_polls = 0;
#endif
			if (likely(res2 >= 0)) {
/*  New msg complete, or first chunk arrived and handler must run.  */
				spin_unlock(&gamma_rxsamenode_lock);
				return res2;
			}
/*  res2 can be:  -GAMMA_RX_INCOMPLETE, -GAMMA_ERR_MUSTYIELD  */
		}

		spin_unlock(&gamma_rxsamenode_lock);

skip_intra:

#endif

		if (res == -GAMMA_ERR_MUSTYIELD) return res;
#ifdef MORE_INSTANCES_ON_SAME_NODE
		if (res2 == -GAMMA_ERR_MUSTYIELD) return res2;
#endif

		gamma_time(b);
		if (gamma_time_diff_ticks(b,a) >= REPEAT_POLL*CLOCK) break;
	}

#ifdef SEQCHECK
	if (++how_many_polls >= max_polls) {
		how_many_polls = 0;

DEBUG_L(printg(
"sys_gamma_poll():  no pkts for a long time.  Testing frame seq #s...\n"
));

DEBUG_F({
 int i;
 for (i=0;i<total_num_nodes;i++)
	printg("%d ",atomic_read(&local_credit[i]));
 printg("\n");
 for (i=0;i<total_num_nodes;i++)
	 printg("%d ",rx_seq[i]);
 printg("\n");
})

		spin_lock(&gamma_tx_lock);

/*  Ask everybody to spread their frame numbering tables.  */
		gamma_kernel_send(BROADCAST,DONT_MIND,TEST,NULL,0,0);
#if 0
/*  Transmit the local frame numbering table (both unicast and bcast).  */
 for (
  i = (phys_whoami + 1) % total_num_nodes;
  i != phys_whoami;
  i = (i + 1) % total_num_nodes
 ) {
  unsigned long frame_no;

  frame_no = tx_tmp_ring[i].tx_seq;
  gamma_kernel_send(i,DONT_MIND,ACK,&frame_no,sizeof(frame_no),0);
  frame_no = tx_tmp_ring[phys_whoami].tx_seq;
  gamma_kernel_send(i,DONT_MIND,ACKB,&frame_no,sizeof(frame_no),0);
 }
#endif

		spin_unlock(&gamma_tx_lock);
	}
#endif

	return -GAMMA_RX_INCOMPLETE;

}


long sys_gamma_poll_samenode
(void)
#ifdef MORE_INSTANCES_ON_SAME_NODE
{

	int res = -GAMMA_RX_INCOMPLETE;


/*  Check for intra-node messages  */

	spin_lock(&gamma_rxsamenode_lock);
	if (shared_ring.head != shared_ring.tail)
		res = rx_gamma_samenode();
	spin_unlock(&gamma_rxsamenode_lock);
	return res;

}
#else
{
	return 0;
}
#endif


long sys_gamma_poll_nowait
(void)
{

	int res  = -GAMMA_RX_INCOMPLETE;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res2 = -GAMMA_RX_INCOMPLETE;
#endif
#ifdef MORE_NICS_SAMENODE_ROUTING
	int i;
	poll_common_decl; 
#else
	poll_common_decl_and_init;
#endif


 /*  Check out the NIC rx ring (inter-node msgs)  */

	if (!spin_trylock(&gamma_rx_lock)) goto skip_inter;

#ifdef MORE_NICS_SAMENODE_ROUTING
	for (i=0; i < num_gamma_devices; i++) {
		rx_read_pkt_info_light_arg(i);
		if (!(rx_err_pktnotavail())) {
			res = rx_gamma(gamma_devices[i]);

#else
	{
		rx_read_pkt_info_light();
		if (!rx_err_pktnotavail()) 
		{
			res = rx_gamma(gamma_devices[0]);
#endif
			clear_rx_irqs();
#ifdef SEQCHECK
			how_many_polls = 0;
#endif
			if (res >= 0) {
  /*  New msg complete, or first chunk arrived and handler must run.  */
				spin_unlock(&gamma_rx_lock);
				return res;
			}
		}
	}

	spin_unlock(&gamma_rx_lock);

skip_inter:

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*  Check for intra-node messages  */

	if (!spin_trylock(&gamma_rxsamenode_lock)) goto skip_intra;

	if (shared_ring.head != shared_ring.tail) {
		res2 = rx_gamma_samenode();
#ifdef SEQCHECK
		how_many_polls = 0;
#endif
		if (res2 >= 0) {
/*  New msg complete, or first chunk arrived and handler must run.  */
			spin_unlock(&gamma_rxsamenode_lock);
			return res2;
		}
	}

	spin_unlock(&gamma_rxsamenode_lock);

skip_intra:

#endif

	if (res == -GAMMA_ERR_MUSTYIELD) return res;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (res2 == -GAMMA_ERR_MUSTYIELD) return res2;
#endif

#ifdef SEQCHECK
	if (++how_many_polls >= (max_polls*2000)) {
			/*
			 *  "2000":  Adjust waiting time to take into account
			 *  that this routine does not spin over Rx (it is a
			 *  "one time only" check) and so returns quickly.
			 */
		how_many_polls = 0;

DEBUG_L(printg(
"sys_gamma_poll_nowait():  no pkts for a long time.  Testing frame seq #s...\n"
));

		spin_lock(&gamma_tx_lock);

/*  Ask everybody to spread their frame numbering tables.  */
		gamma_kernel_send(BROADCAST,DONT_MIND,TEST,NULL,0,0);
#if 0
/*  Transmit the local frame numbering table (both unicast and bcast).  */
 for (
  i = (phys_whoami + 1) % total_num_nodes;
  i != phys_whoami;
  i = (i + 1) % total_num_nodes
 ) {
  unsigned long frame_no;

  frame_no = tx_tmp_ring[i].tx_seq;
  gamma_kernel_send(i,DONT_MIND,ACK,&frame_no,sizeof(frame_no),0);
  frame_no = tx_tmp_ring[phys_whoami].tx_seq;
  gamma_kernel_send(i,DONT_MIND,ACKB,&frame_no,sizeof(frame_no),0);
 }
#endif

		spin_unlock(&gamma_tx_lock);
	}
#endif

	return -GAMMA_RX_INCOMPLETE;

}


/************************************************************************
 *	Miscellaneous routines						*
 ************************************************************************/


/*  Disable IRQ-driven message delivery  */
long sys_gamma_suspend_irq
(void)
{
	spin_lock(&gamma_susp_irq_count_lock);
	if (!(gamma_susp_irq_count++))
		rx_suspend_irq();
	spin_unlock(&gamma_susp_irq_count_lock);
	return 0;
}


/*  Enable IRQ-driven message delivery  */
long sys_gamma_resume_irq
(void)
{
	spin_lock(&gamma_susp_irq_count_lock);
	if (gamma_susp_irq_count)
		if (!(--gamma_susp_irq_count))
			rx_resume_irq();
	spin_unlock(&gamma_susp_irq_count_lock);
	return 0;
}


/*
 *  Print an integer value to "dmesg".
 */
long sys_gamma_printk
(long val)
{
	printg("sys_gamma_printk(): process %s (pid %d): %ld (0x%016lx)\n",
		current->comm,current->pid,val,val
	);
	return 0;
}


/*
 *  Set, reset, get the initial credit size.
 */
long sys_gamma_setcredit
(unsigned long size)
{
	int i;

	credit_size = size;
	for (i=0;i<total_num_nodes;i++) {
		initial_credit[i] = size;
		credit_ceil[i] = size;
		credit_growth_cnt[i] = 0;
		atomic_set(&local_credit[i],initial_credit[i]);
	}
	return 0;
}

long sys_gamma_resetcredit
(void)
{
	int i;

	credit_size = MAX_CREDIT_SIZE;
	for (i=0;i<total_num_nodes;i++) {
		initial_credit[i] = MAX_CREDIT_SIZE;
		credit_ceil[i] = MAX_CREDIT_SIZE;
		credit_growth_cnt[i] = 0;
		atomic_set(&local_credit[i],initial_credit[i]);
	}
	return 0;
}

long sys_gamma_getcredit
(void)
{
	int i;

	printg("sys_gamma_getcredit(): ");
	for (i=0;i<total_num_nodes;i++) {
		printg("-->%d: %ld",i,initial_credit[i]);
		if (i<total_num_nodes-1)
			printg(", ");
	}
	printg("\n");
/*
 *  We only return the credit size towards node 0.  The complete picture has
 *  been printed by the statements above.
 */
	return initial_credit[0];
}


/*
 *  Set and reset the "max_polls" variables (threshold which triggers the
 *  collective check of frame numbering and credits.
 */
long sys_gamma_setmaxpolls
(unsigned long n)
{
	max_polls = n;
	base_max_polls = n;
	return 0;
}

long sys_gamma_resetmaxpolls
(void)
{
	max_polls = MAX_POLLS;
	base_max_polls = MAX_POLLS;
	return 0;
}

long sys_gamma_getmaxpolls
(void)
{
	return base_max_polls;
}


/************************************************************************
 *	Local routines							*
 ************************************************************************/


#define low_watermark_pt2pt(N) \
(LOW_WATERMARK < initial_credit[N] ? LOW_WATERMARK : initial_credit[N])

static inline int is_low_watermark
(unsigned long node, unsigned long credit, unsigned long curr_credit)
{
	int i;

	if (node != BROADCAST)
		return (curr_credit == low_watermark_pt2pt(node));
	for (i=0; i<total_num_nodes; i++)
		if (i != phys_whoami)
			if (
				(atomic_read(&local_credit[i])
				 - credit + curr_credit)
				== low_watermark_pt2pt(i)
			)
				return 1;
	return 0;
}


/*
 *  Other support routines.
 */
#include "sys_gamma_connect_port.c"
#include "sys_gamma_vm.c"
#include "sys_gamma_mem.c"
#include "sys_gamma_timers.c"
